"""
implementations of the different dtypes supported in heat
"""

from __future__ import annotations

import builtins
import collections
from functools import cache
import numpy as np
import torch

from . import communication
from . import devices
from . import factories
from . import _operations
from . import sanitation

from typing import Type, Union, Tuple, Any, Iterable, Optional

__all__ = [
    "datatype",
    "number",
    "integer",
    "signedinteger",
    "unsignedinteger",
    "bool",
    "bool_",
    "floating",
    "int8",
    "byte",
    "int16",
    "short",
    "int32",
    "int",
    "int64",
    "long",
    "uint8",
    "ubyte",
    "float16",
    "half",
    "float32",
    "float",
    "float_",
    "float64",
    "double",
    "flexible",
    "can_cast",
    "canonical_heat_type",
    "heat_type_is_exact",
    "heat_type_is_inexact",
    "heat_type_is_realfloating",
    "heat_type_is_complexfloating",
    "iscomplex",
    "isreal",
    "issubdtype",
    "heat_type_of",
    "promote_types",
    "result_type",
    "complex64",
    "cfloat",
    "csingle",
    "complex128",
    "cdouble",
]
__api__ = []


class datatype:
    """
    Defines the basic heat data types in the hierarchy as shown below. Design inspired by the Python package numpy.
    As part of the type-hierarchy: xx -- is bit-width \n
        - generic \n
            - bool, bool_ (kind=?) \n
            - number \n
                - integer \n
                    - signedinteger (intxx)(kind=b, i) \n
                        - int8, byte \n
                        - int16, short \n
                        - int32, int \n
                        - int64, long
                    - unsignedinteger (uintxx)(kind=B, u) \n
                        - uint8, ubyte
                - floating (floatxx) (kind=f) \n
                    - float16, half \n
                    - float32, float, float_ \n
                    - float64, double (double)
            - flexible (currently unused, placeholder for characters) \n
    """

    _can_be_cast_safely_to = []
    _can_be_cast_intuitively_to = []  # types that cannot be cast safely, but only intuitively
    mappings = []  # things that map to this datatype

    def __new__(
        cls,
        *value,
        device: Optional[Union[str, devices.Device]] = None,
        comm: Optional[communication.Communication] = None,
    ) -> dndarray.DNDarray:
        """
        Create a new DNDarray. See :func:`ht.array <heat.core.factories.array>` for more info on general
        DNDarray creation.

        Parameters
        ----------
        value: array_like
            The values for the DNDarray which will be created
        device: devices.Device
            The device on which to place the created DNDarray
        comm: communication.Communication
            The MPI communication object to use in distribution and further operations
        """
        torch_type = cls.torch_type()
        if torch_type is NotImplemented:
            raise TypeError(f"cannot create '{cls}' instances")

        value_count = len(value)

        # sanitize the distributed processing flags
        comm = communication.sanitize_comm(comm)
        device = devices.sanitize_device(device)

        # check whether there are too many arguments
        if value_count >= 2:
            raise TypeError(f"function takes at most 1 argument ({value_count} given)")
        elif value_count == 0:
            value = ((0,),)

        # otherwise, attempt to create a torch tensor of given type
        try:
            array = value[0]._DNDarray__array.type(torch_type)
            return dndarray.DNDarray(
                array,
                gshape=value[0].shape,
                dtype=cls,
                split=value[0].split,
                comm=comm,
                device=device,
                balanced=value[0].balanced,
            )
        except AttributeError:
            # this is the case of that the first/only element of value is not a DNDarray
            array = torch.tensor(*value, dtype=torch_type, device=device.torch_device)
        except TypeError as exception:
            # re-raise the exception to be consistent with numpy's exception interface
            raise ValueError(str(exception))

        return dndarray.DNDarray(
            array, tuple(array.shape), cls, split=None, device=device, comm=comm, balanced=True
        )

    @classmethod
    def torch_type(cls) -> NotImplemented:
        """
        Torch Datatype
        """
        return NotImplemented

    @classmethod
    def char(cls) -> NotImplemented:
        """
        Datatype short-hand name
        """
        return NotImplemented

    @classmethod
    def can_cast_to(
        cls,
        to,
        casting: str = "intuitive",
    ) -> bool:
        """
        Returns True if cast between this datatype and `to`-datatype can occur according to the casting rule. If `to` is a scalar or array
        scalar, also returns True if the scalar value can be cast without overflow or truncation to an integer.

        Parameters
        ----------
        to : Union[str, Type[datatype], Any]
            Target type to cast to.
        casting: str, optional
            options: {"no", "safe", "same_kind", "unsafe", "intuitive"}, optional
            Controls the way the cast is evaluated
                * "no" the types may not be cast, i.e. they need to be identical
                * "safe" allows only casts that can preserve values with complete precision
                * "same_kind" safe casts are possible and down_casts within the same type family, e.g. int32 -> int8
                * "unsafe" means any conversion can be performed, i.e. this casting is always possible
                * "intuitive" allows all of the casts of safe plus casting from int32 to float32


        Raises
        ------
        TypeError
            If the types are not understood or casting is not a string
        ValueError
            If the casting rule is not understood

        Examples
        --------
        >>> ht.int32.can_cast_to(ht.int64)
        True
        >>> ht.int64.can_cast_to(ht.float64)
        True
        >>> ht.int16.can_cast_to(ht.int8)
        False
        >>> ht.int16.can_cast_to(1)
        True
        >>> ht.int16.can_cast_to("f4")
        True
        >>> ht.int64.can_cast_to("i2", casting="safe")
        False
        """
        if not isinstance(casting, str):
            raise TypeError(f"expected string, found {type(casting)}")

        try:
            type_to = canonical_heat_type(to)
        except TypeError:
            type_to = heat_type_of(to)

        name_type_to = type_to.__name__

        if casting == "no":
            return type_to == cls
        elif casting == "unsafe":
            return True
        elif casting == "same_kind":
            for base_type in [floating, integer, complex, bool]:
                if issubdtype(type_to, base_type) and issubdtype(cls, base_type):
                    return True
            return False
        elif casting == "intuitive":
            return (
                name_type_to in cls._can_be_cast_safely_to
                or name_type_to in cls._can_be_cast_intuitively_to
            )
        elif casting == "safe":
            return name_type_to in cls._can_be_cast_safely_to
        else:
            cast_kinds = ["no", "unsafe", "safe", "same_kind", "intuitive"]
            raise ValueError(f"casting must be one of {cast_kinds}, not {casting}")


class bool(datatype):
    """
    The boolean datatype in Heat
    """

    _can_be_cast_safely_to = [
        "bool",
        "uint8",
        "int8",
        "int16",
        "int32",
        "int64",
        "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    mappings = ["bool_", "?", np.bool_, torch.bool, builtins.bool]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.bool

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "u1"


class number(datatype):
    """
    The general number datatype. Integer and Float classes will inherit from this.
    """

    pass


class integer(number):
    """
    The general integer datatype. Specific integer classes inherit from this.
    """

    pass


class signedinteger(integer):
    """
    The general signed integer datatype.
    """

    pass


class int8(signedinteger):
    """
    8 bit signed integer datatype
    """

    _can_be_cast_safely_to = [
        "int8",
        "int16",
        "int32",
        "int64",
        "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    mappings = ["b", "i1", np.int8, torch.int8]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int8

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i1"


class int16(signedinteger):
    """
    16 bit signed integer datatype
    """

    _can_be_cast_safely_to = [
        "int16",
        "int32",
        "int64",
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    _can_be_cast_intuitively_to = ["float16"]
    mappings = ["h", "i2", np.int16, torch.int16]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int16

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i2"


class int32(signedinteger):
    """
    32 bit signed integer datatype
    """

    _can_be_cast_safely_to = ["int32", "int64", "float64", "complex128"]
    _can_be_cast_intuitively_to = ["float32", "complex64"]
    mappings = ["i", "i4", np.int32, torch.int32, builtins.int]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int32

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i4"


class int64(signedinteger):
    """
    64 bit signed integer datatype
    """

    _can_be_cast_safely_to = ["int64", "float64", "complex128"]
    mappings = ["l", "i8", np.int64, torch.int64]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.int64

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "i8"


class unsignedinteger(integer):
    """
    The general unsigned integer datatype
    """

    pass


class uint8(unsignedinteger):
    """
    8 bit unsigned integer datatype
    """

    _can_be_cast_safely_to = [
        "uint8",
        "int16",
        "int32",
        "int64",
        "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    mappings = ["B", "u", "u1", np.uint8, torch.uint8]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.uint8

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "u1"


class floating(number):
    """
    The general floating point datatype class.
    """

    pass


class float16(floating):
    """
    The 16 bit floating point datatype
    """

    _can_be_cast_safely_to = [
        "float16",
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    mappings = ["f2", np.float16, torch.float16]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.float16

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "f2"


class float32(floating):
    """
    The 32 bit floating point datatype
    """

    _can_be_cast_safely_to = [
        "float32",
        "float64",
        "complex64",
        "complex128",
    ]
    mappings = ["f", "f4", np.float32, torch.float32, builtins.float]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatype
        """
        return torch.float32

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "f4"


class float64(floating):
    """
    The 64 bit floating point datatype
    """

    _can_be_cast_safely_to = [
        "float64",
        "complex128",
    ]
    mappings = ["d", "f8", np.float64, torch.float64]

    @classmethod
    def torch_type(cls) -> torch.dtype:
        """
        Torch Datatye
        """
        return torch.float64

    @classmethod
    def char(cls) -> str:
        """
        Datatype short-hand name
        """
        return "f8"


class flexible(datatype):
    """
    The general flexible datatype. Currently unused, placeholder for characters
    """

    pass


class complex(number):
    """
    The general complex datatype class.
    """


class complex64(complex):
    """
    The complex 64 bit datatype. Both real and imaginary are 32 bit floating point
    """

    _can_be_cast_safely_to = [
        "complex64",
        "complex128",
    ]
    mappings = ["F", "c8", np.complex64, torch.complex64, builtins.complex]

    @classmethod
    def torch_type(cls):
        """
        Torch Datatype
        """
        return torch.complex64

    @classmethod
    def char(cls):
        """
        Datatype short-hand name
        """
        return "c8"


class complex128(complex):
    """
    The complex 128 bit datatype. Both real and imaginary are 64 bit floating point
    """

    _can_be_cast_safely_to = [
        "complex128",
    ]
    mappings = ["D", "c16", np.complex128, torch.complex128]

    @classmethod
    def torch_type(cls):
        """
        Torch Datatype
        """
        return torch.complex128

    @classmethod
    def char(cls):
        """
        Datatype short-hand name
        """
        return "c16"


# definition of aliases
bool_ = bool
ubyte = uint8
byte = int8
short = int16
int = int32
int_ = int32
long = int64
half = float16
float = float32
float_ = float32
double = float64
cfloat = complex64
csingle = complex64
cdouble = complex128


@cache
def __get_all_heat_types():
    def all_subclasses(cls):
        return set(cls.__subclasses__()).union(
            {g for s in cls.__subclasses__() for g in all_subclasses(s)}
        )

    type_related_classes = all_subclasses(datatype)
    return [me for me in type_related_classes if len(me.__subclasses__()) == 0]


def canonical_heat_type(a_type: Union[str, Type[datatype], Any]) -> Type[datatype]:
    """
    Canonicalize the builtin Python type, type string or HeAT type into a canonical HeAT type.

    Parameters
    ----------
    a_type : type, str, datatype
        A description for the type. It may be a Python builtin type, string or an HeAT type already.
        In the three former cases the according mapped type is looked up, in the latter the type is simply returned.

    Raises
    ------
    TypeError
        If the type cannot be converted.
    """
    # already a heat type
    try:
        if issubclass(a_type, datatype):
            return a_type
    except TypeError:
        pass

    # extract type of numpy.dtype
    a_type = getattr(a_type, "type", a_type)

    # try to look the corresponding type up
    all_types = __get_all_heat_types()
    for type_ in all_types:
        try:
            if a_type in type_.mappings:
                return type_
        except ValueError:
            # this particular type_ cannot interpret a_type, try next candidate
            continue
    raise TypeError(f"data type {a_type} is not understood")


def heat_type_is_exact(ht_dtype: Type[datatype]) -> bool:
    """
    Check if HeAT type is an exact type, i.e an integer type. True if ht_dtype is an integer, False otherwise

    Parameters
    ----------
    ht_dtype: Type[datatype]
        HeAT type to check
    """
    return issubdtype(ht_dtype, integer)


def heat_type_is_inexact(ht_dtype: Type[datatype]) -> bool:
    """
    Check if HeAT type is an inexact type, i.e floating point type. True if ht_dtype is a float or complex, False otherwise

    Parameters
    ----------
    ht_dtype: Type[datatype]
        HeAT type to check
    """
    return issubdtype(ht_dtype, floating) or issubdtype(ht_dtype, complex)


def heat_type_is_realfloating(ht_dtype: Type[datatype]) -> bool:
    """
    Check if Heat type is a real floating point number, i.e float32 or float64

    Parameters
    ----------
    ht_dtype: Type[datatype]
        Heat type to check

    Returns
    -------
    out: bool
        True if ht_dtype is a real float, False otherwise
    """
    return issubclass(ht_dtype, floating)


def heat_type_is_complexfloating(ht_dtype: Type[datatype]) -> bool:
    """
    Check if Heat type is a complex floating point number, i.e complex64

    Parameters
    ----------
    ht_dtype: ht.dtype
        HeAT type to check

    Returns
    -------
    out: bool
        True if ht_dtype is a complex float, False otherwise
    """
    try:
        ht_dtype = canonical_heat_type(ht_dtype)
    except TypeError:
        ht_dtype = heat_type_of(ht_dtype)
    return issubclass(ht_dtype, complex)


def heat_type_of(
    obj: Union[str, Type[datatype], Any, Iterable[str, Type[datatype], Any]],
) -> Type[datatype]:
    """
    Returns the corresponding HeAT data type of given object, i.e. scalar, array or iterable. Attempts to determine the
    canonical data type based on the following priority list:
        1. dtype property
        2. type(obj)
        3. type(obj[0])

    Parameters
    ----------
    obj : scalar or DNDarray or iterable
        The object for which to infer the type.

    Raises
    ------
    TypeError
        If the object's type cannot be inferred.
    """
    # attempt to access the dtype property
    try:
        return canonical_heat_type(obj.dtype)
    except (AttributeError, TypeError):
        pass

    # attempt type of object itself
    try:
        return canonical_heat_type(type(obj))
    except TypeError:
        pass

    # last resort, type of the object at first position
    try:
        return canonical_heat_type(type(obj[0]))
    except (KeyError, IndexError, TypeError):
        raise TypeError(f"data type of {obj} is not understood")


def can_cast(
    from_: Union[str, Type[datatype], Any],
    to: Union[str, Type[datatype], Any],
    casting: str = "intuitive",
) -> bool:
    """
    Returns True if cast between data types can occur according to the casting rule. If from is a scalar or array
    scalar, also returns True if the scalar value can be cast without overflow or truncation to an integer.

    Parameters
    ----------
    from_ : Union[str, Type[datatype], Any]
        Scalar, data type or type specifier to cast from.
    to : Union[str, Type[datatype], Any]
        Target type to cast to.
    casting: str, optional
        options: {"no", "safe", "same_kind", "unsafe", "intuitive"}, optional
        Controls the way the cast is evaluated
            * "no" the types may not be cast, i.e. they need to be identical
            * "safe" allows only casts that can preserve values with complete precision
            * "same_kind" safe casts are possible and down_casts within the same type family, e.g. int32 -> int8
            * "unsafe" means any conversion can be performed, i.e. this casting is always possible
            * "intuitive" allows all of the casts of safe plus casting from int32 to float32


    Raises
    ------
    TypeError
        If the types are not understood

    Examples
    --------
    >>> ht.can_cast(ht.int32, ht.int64)
    True
    >>> ht.can_cast(ht.int64, ht.float64)
    True
    >>> ht.can_cast(ht.int16, ht.int8)
    False
    >>> ht.can_cast(1, ht.float64)
    True
    >>> ht.can_cast(2.0e200, "u1")
    False
    >>> ht.can_cast("i8", "i4", "no")
    False
    >>> ht.can_cast("i8", "i4", "safe")
    False
    >>> ht.can_cast("i8", "i4", "same_kind")
    True
    >>> ht.can_cast("i8", "i4", "unsafe")
    True
    """
    # obtain the types codes of the canonical HeAT types
    try:
        type_from = canonical_heat_type(from_)
    except TypeError:
        type_from = heat_type_of(from_)

    try:
        type_to = canonical_heat_type(to)
    except TypeError:
        type_to = heat_type_of(to)

    return type_from.can_cast_to(type_to, casting=casting)


def iscomplex(x: dndarray.DNDarray) -> dndarray.DNDarray:
    """
    Test element-wise if input is complex.

    Parameters
    ----------
    x : DNDarray
        The input DNDarray

    Examples
    --------
    >>> ht.iscomplex(ht.array([1 + 1j, 1]))
    DNDarray([ True, False], dtype=ht.bool, device=cpu:0, split=None)
    """
    sanitation.sanitize_in(x)

    if heat_type_is_complexfloating(canonical_heat_type(x.dtype)):
        return x.imag != 0
    else:
        return factories.zeros(x.shape, bool, split=x.split, device=x.device, comm=x.comm)


def isreal(x: dndarray.DNDarray) -> dndarray.DNDarray:
    """
    Test element-wise if input is real-valued.

    Parameters
    ----------
    x : DNDarray
        The input DNDarray

    Examples
    --------
    >>> ht.iscomplex(ht.array([1 + 1j, 1]))
    DNDarray([ True, False], dtype=ht.bool, device=cpu:0, split=None)
    """
    return _operations.__local_op(torch.isreal, x, None, no_cast=True)


def issubdtype(
    arg1: Union[str, Type[datatype], Any], arg2: Union[str, Type[datatype], Any]
) -> builtins.bool:
    """
    Returns True if first argument is a typecode lower/equal in type hierarchy.

    Parameters
    ----------
    arg1 : type, str, ht.dtype
        A description representing the type. It may be a a Python builtin type, string or an HeAT type already.
    arg2 : type, str, ht.dtype
        A description representing the type. It may be a a Python builtin type, string or an HeAT type already.


    Examples
    --------
    >>> ints = ht.array([1, 2, 3], dtype=ht.int32)
    >>> ht.issubdtype(ints.dtype, ht.integer)
    True
    >>> ht.issubdype(ints.dtype, ht.floating)
    False
    >>> ht.issubdtype(ht.float64, ht.float32)
    False
    >>> ht.issubdtype("i", ht.integer)
    True
    """
    # Assure that each argument is a ht.dtype
    arg1 = canonical_heat_type(arg1)
    arg2 = canonical_heat_type(arg2)

    return issubclass(arg1, arg2)


def promote_types(
    type1: Union[str, Type[datatype], Any], type2: Union[str, Type[datatype], Any]
) -> Type[datatype]:
    """
    Returns the data type with the smallest size and smallest scalar kind to which both ``type1`` and ``type2`` may be
    intuitively cast to, where intuitive casting refers to maintaining the same bit length if possible. This function
    is symmetric.

    Parameters
    ----------
    type1 : type or str or datatype
        type of first operand
    type2 : type or str or datatype
        type of second operand

    Examples
    --------
    >>> ht.promote_types(ht.uint8, ht.uint8)
    <class 'heat.core.types.uint8'>
    >>> ht.promote_types(ht.int32, ht.float32)
    <class 'heat.core.types.float32'>
    >>> ht.promote_types(ht.int8, ht.uint8)
    <class 'heat.core.types.int16'>
    >>> ht.promote_types("i8", "f4")
    <class 'heat.core.types.float64'>
    """
    heat_type1 = canonical_heat_type(type1)
    heat_type2 = canonical_heat_type(type2)

    # get all options the types can be cast to intuitively
    options_1 = heat_type1._can_be_cast_safely_to + heat_type1._can_be_cast_intuitively_to
    options_2 = heat_type2._can_be_cast_safely_to + heat_type2._can_be_cast_intuitively_to

    # get the types both types can be cast to intuitively
    shared = [me for me in options_1 if me in options_2]
    if len(shared) == 0:
        raise Exception(
            f"Cannot find type that {heat_type1} and {heat_type2} can both be cast to intuitively. Please report this bug to the developers."
        )

    # get the sizes in bits
    sizes = [np.int32(me[len(me.rstrip("0123456789")) :]) if me != "bool" else 1 for me in shared]

    # return shared datatype with the smallest size
    return [me for me in __get_all_heat_types() if me.__name__ == shared[np.argmin(sizes)]][0]


def result_type(
    *arrays_and_types: Tuple[Union[dndarray.DNDarray, Type[datatype], Any]],
) -> Type[datatype]:
    """
    Returns the data type that results from type promotions rules performed in an arithmetic operation.

    Parameters
    ----------
    arrays_and_types: List of arrays and types
        Input arrays, types or numbers of the operation.

    Examples
    --------
    >>> ht.result_type(ht.array([1], dtype=ht.int32), 1)
    ht.int32
    >>> ht.result_type(ht.float32, ht.array(1, dtype=ht.int8))
    ht.float32
    >>> ht.result_type("i8", "f4")
    ht.float64
    """

    def result_type_rec(*arrays_and_types):
        # derive type and set precedence (lower number, higher precedence)
        arg = arrays_and_types[0]

        try:
            # array / tensor
            if isinstance(arg, np.ndarray):
                type1 = canonical_heat_type(arg.dtype.char)
            else:
                type1 = canonical_heat_type(arg.dtype)

            if len(arg.shape) > 0:
                prec1 = 0  # array
            else:
                prec1 = 2  # scalar
        except (AttributeError, TypeError):
            try:
                # type
                if isinstance(arg, np.dtype):
                    arg = arg.char
                type1 = canonical_heat_type(arg)
                prec1 = 1
            except TypeError:
                # type instance
                type1 = canonical_heat_type(type(arg))
                prec1 = 3

        # multiple arguments
        if len(arrays_and_types) > 1:
            type2, prec2 = result_type_rec(*arrays_and_types[1:])

            # fast check same type
            if type1 == type2:
                return type1, min(prec1, prec2)
            # fast check same precedence
            if prec1 == prec2:
                return promote_types(type1, type2), prec1

            # check if parent type is identical and decide by precedence
            if can_cast(type1, type2, casting="same_kind"):
                if prec1 < prec2:
                    return type1, min(prec1, prec2)
                else:
                    return type2, min(prec1, prec2)

            # different parent type: bool < int < float < complex
            dtype_priority_order = [bool, integer, floating, complex]
            try:
                dtype_priority1 = [
                    issubdtype(type1, dtype) for dtype in dtype_priority_order
                ].index(True)
            except ValueError:
                return type2, prec2
            try:
                dtype_priority2 = [
                    issubdtype(type2, dtype) for dtype in dtype_priority_order
                ].index(True)
            except ValueError:
                return type1, prec1

            if dtype_priority1 < dtype_priority2:
                return type2, min(prec1, prec2)
            else:
                return type1, min(prec1, prec2)

        # single argument
        return type1, prec1

    return result_type_rec(*arrays_and_types)[0]


class finfo:
    """
    Class describing machine limits (bit representation) of floating point types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    eps : float
        The smallest representable positive number such that
        ``1.0 + eps != 1.0``.  Type of ``eps`` is an appropriate floating
        point type.
    max : float
        The largest representable number.
    min : float
        The smallest representable number, typically ``-max``.
    tiny : float
        The smallest positive usable number.  Type of ``tiny`` is an
        appropriate floating point type.

    Parameters
    ----------
    dtype : datatype
        Kind of floating point data-type about which to get information.

    Examples
    --------
    >>> import heat as ht
    >>> info = ht.types.finfo(ht.float32)
    >>> info.bits
    32
    >>> info.eps
    1.1920928955078125e-07
    """

    def __new__(cls, dtype: Type[datatype]):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError):
            # If given type is not heat type
            pass

        if not heat_type_is_inexact(dtype):
            raise TypeError(f"Data type {dtype} not inexact, not supported")

        return super(finfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype: Type[datatype]):
        _torch_finfo = torch.finfo(dtype.torch_type())
        for word in ["bits", "eps", "max", "min", "tiny"]:
            setattr(self, word, getattr(_torch_finfo, word))

        return self


class iinfo:
    """
    Class describing machine limits (bit representation) of integer types.

    Attributes
    ----------
    bits : int
        The number of bits occupied by the type.
    max : float
        The largest representable number.
    min : float
        The smallest representable number, typically ``-max``.

    Parameters
    ----------
    dtype : datatype
        Kind of floating point data-type about which to get information.

    Examples
    --------
    >>> import heat as ht
    >>> info = ht.types.iinfo(ht.int32)
    >>> info.bits
    32
    """

    def __new__(cls, dtype: Type[datatype]):
        try:
            dtype = heat_type_of(dtype)
        except (KeyError, IndexError, TypeError):
            # If given type is not heat type
            pass

        if not heat_type_is_exact(dtype):
            raise TypeError(f"Data type {dtype} not exact, not supported")

        return super(iinfo, cls).__new__(cls)._init(dtype)

    def _init(self, dtype: Type[datatype]):
        _torch_iinfo = torch.iinfo(dtype.torch_type())
        for word in ["bits", "min", "max"]:
            setattr(self, word, getattr(_torch_iinfo, word))

        return self


# dndarray is imported at the very end to break circular dependency
from . import dndarray



from typing import Union
from ._dtypes import (
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
    float32,
    float64,
    bool,
)

from heat.core.devices import Device, cpu

try:
    from heat.core.devices import gpu
except ImportError:
    pass

Dtype = Union[
    int8,
    int16,
    int32,
    int64,
    uint8,
    # uint16,
    # uint32,
    # uint64,
    float32,
    float64,
    bool,
]



"""
Implements Symmetric Eigenvalue Decomposition
"""

import numpy as np
import collections
import torch
from typing import Type, Callable, Dict, Any, TypeVar, Union, Tuple

from ..dndarray import DNDarray
from .. import factories
from .. import types
from ..linalg import matrix_norm, vector_norm, matmul, qr, polar
from ..indexing import where
from ..random import randn
from ..devices import Device
from ..manipulations import vstack, hstack, concatenate, diag, balance
from .. import statistics
from mpi4py import MPI
from ..sanitation import sanitize_in_nd_realfloating


__all__ = ["eigh"]


def _subspaceiteration(
    A: DNDarray,
    C: DNDarray,
    silent: bool = True,
    safetyparam: int = 3,
    maxit: int = None,
    tol: float = None,
    depth: int = 0,
) -> DNDarray:
    """
    Auxiliary function that implements the subspace iteration as required for symmetric eigenvalue decomposition
    via polar decomposition; cf. Ref. 2 below. The algorithm for subspace iteration itself is taken from Ref. 1,
    Algorithm 3 in Sect. 5.1.

    Given a symmetric matrix ``A`` and a matrix ``C`` that is the orthogonal projection onto an invariant
    subspace of A, this function computes and returns an orthogonal matrix ``Q`` such that Q = [V_1 V_2] with
    C = V_1 V_1.T. Moreover, the dimension of the invariant subspace, i.e., the number of column of V_1 is
    returned as well.

    References
    ----------
    1.  Nakatsukasa, Y., & Higham, N. J. (2013). Stable and efficient spectral divide and conquer algorithms for
        Hermitian eigenproblems. SIAM Journal on Scientific Computing, 35(3).
    2.  Nakatsukasa, Y., & Freund, R. W. (2016). Computing fundamental matrix decompositions accurately via the
        matrix sign function in two iterations: The power of Zolotarev's functions. SIAM Review, 58(3).
    """
    # set parameters for convergence
    if A.dtype == types.float64:
        maxit = 3 if maxit is None else maxit
        tol = 1e-8 if tol is None else tol
    elif A.dtype == types.float32:
        maxit = 6 if maxit is None else maxit
        tol = 1e-4 if tol is None else tol
    else:
        raise TypeError(
            f"Input DNDarray must be of data type float32 or float64, but is of type {A.dtype}."
        )

    Anorm = matrix_norm(A, ord="fro")

    # this initialization is proposed in Ref. 1, Sect. 5.1
    k = int(round(matrix_norm(C, ord="fro").item() ** 2))
    columnnorms = vector_norm(C, axis=0)
    idx = where(
        columnnorms
        >= factories.ones(
            columnnorms.shape,
            comm=columnnorms.comm,
            split=columnnorms.split,
            device=columnnorms.device,
        )
        * statistics.percentile(columnnorms, 100.0 * (1 - (k + safetyparam) / columnnorms.shape[0]))
    )
    X = C[:, idx].balance()

    # actual subspace iteration
    it = 1
    while it < maxit + 1:
        # enrich X by additional random columns to get a full orthonormal basis by QR
        if X.shape[0] > X.shape[1]:
            X = hstack(
                [
                    X,
                    randn(
                        X.shape[0],
                        X.shape[0] - X.shape[1],
                        dtype=X.dtype,
                        device=X.device,
                        comm=X.comm,
                        split=X.split,
                    ),
                ]
            )
        Q, _ = qr(X)
        Q_k = Q[:, :k].balance()
        Q_k_orth = Q[:, k:].balance()
        E = (Q_k_orth.T @ A) @ Q_k
        Enorm = matrix_norm(E, ord="fro")
        if Enorm / Anorm < tol:
            # exit if success
            if A.comm.rank == 0 and not silent:
                print("\t" * depth + f"            Number of subspace iterations: {it}")
            return Q, k
        # else go on with iteration
        X = C @ Q_k
        it += 1
    # warning if the iteration did not converge within the maximum number of iterations
    if A.comm.rank == 0 and not silent:
        print(
            "\t" * depth
            + f"            Subspace iteration did not converge in {maxit} iterations. \n"
            + "\t" * depth
            + f"            It holds ||E||_F/||A||_F = {Enorm / Anorm}, which might impair the accuracy of the result."  # noqa E226
        )
    return Q, k


def _eigh(
    A: DNDarray,
    r: int = None,
    silent: bool = True,
    r_max: int = 8,
    depth: int = 0,
    orig_lsize: int = 0,
) -> Tuple[DNDarray, DNDarray]:
    """
    Auxiliary function for eigh containing the main algorithmic content.
    Inputs are as for the public `eigh`-function, except for:
        `depth`:  an internal variable that is used to track the recursion depth,
        `orig_lsize` an internal variable that is used to propagate the local shapes of the original input matrix
            through the recursions in order to determine when the direct solution of the reduced problems is possible),
        `r`: a hyperparameter for the computation of the polar decomposition via :func:`heat.linalg.polar` which is
            applied multiple times in this function. See the documentation of :func:`heat.linalg.polar` for more details.
            In the actual implementation, this parameter is set to `None` for simplicity.
    """
    n = A.shape[0]
    global_comm = A.comm
    nprocs = global_comm.Get_size()
    rank = global_comm.rank

    # direct solution in torch if the problem is small enough
    if n <= orig_lsize or not A.is_distributed():
        orig_split = A.split
        A.resplit_(None)
        Lambda_loc, Q_loc = torch.linalg.eigh(A.larray)
        Lambda = factories.array(
            torch.flip(Lambda_loc, (0,)), split=0, comm=A.comm, device=A.device
        )
        V = factories.array(torch.flip(Q_loc, (1,)), split=orig_split, comm=A.comm, device=A.device)
        A.resplit_(orig_split)
        return Lambda, V

    if orig_lsize == 0:
        orig_lsize = min(A.lshape_map[:, A.split])

    # now we handle the main case: Zolo-PD is used to reduce the problem to two independent problems
    sigma = statistics.median(diag(A))

    U = polar.polar(
        A
        - sigma * factories.eye((n, n), dtype=A.dtype, device=A.device, comm=A.comm, split=A.split),
        r,
        False,
    )

    V, k = _subspaceiteration(
        A,
        0.5
        * (U + factories.eye((n, n), dtype=A.dtype, device=A.device, comm=A.comm, split=A.split)),
        silent,
        depth,
    )
    A = V.T @ A @ V

    if A.comm.rank == 0 and not silent:
        print(
            "\t" * depth
            + f"At depth {depth}: Zolo-PD(r={'auto' if r is None else r}) on {nprocs} processes reduced symmetric eigenvalue problem of size {n} to"
        )
        print(
            "\t" * depth
            + f"            two independent problems of size {k} and {n - k} respectively."
        )

    # from the "global" A, two independent "local" A's are created
    # the number of processes per local array is roughly proportional to their size with the constraint that
    # each "local" A needs to get at least one process
    nprocs1 = max(1, min(nprocs - 1, round(k / n * nprocs)))
    nprocs2 = nprocs - nprocs1
    new_lshapes = torch.tensor(
        [k // nprocs1 + (i < k % nprocs1) for i in range(nprocs1)]
        + [(n - k) // nprocs2 + (i < (n - k) % nprocs2) for i in range(nprocs2)]
    )
    new_lshape_map = A.lshape_map
    new_lshape_map[:, A.split] = new_lshapes
    A.redistribute_(target_map=new_lshape_map)
    local_comm = A.comm.Split(color=rank < nprocs1, key=rank)
    if A.split == 1:
        A_local = factories.array(
            A.larray[:k, :] if rank < nprocs1 else A.larray[k:, :],
            comm=local_comm,
            is_split=A.split,
        )
    else:
        A_local = factories.array(
            A.larray[:, :k] if rank < nprocs1 else A.larray[:, k:],
            comm=local_comm,
            is_split=A.split,
        )

    Lambda_local, V_local = _eigh(A_local, r, silent, r_max, depth + 1, orig_lsize)

    Lambda = factories.array(Lambda_local.larray, is_split=0, comm=A.comm)
    V_local_larray = V_local.larray
    if A.split == 0:
        if rank < nprocs1:
            V_local_larray = torch.hstack(
                [
                    V_local_larray,
                    torch.zeros(V_local_larray.shape[0], n - k, device=V_local.device.torch_device),
                ]
            )
        else:
            V_local_larray = torch.hstack(
                [
                    torch.zeros(V_local_larray.shape[0], k, device=V_local.device.torch_device),
                    V_local_larray,
                ]
            )
    else:
        if rank < nprocs1:
            V_local_larray = torch.vstack(
                [
                    V_local_larray,
                    torch.zeros(n - k, V_local_larray.shape[1], device=V_local.device.torch_device),
                ]
            )
        else:
            V_local_larray = torch.vstack(
                [
                    torch.zeros(k, V_local_larray.shape[1], device=V_local.device.torch_device),
                    V_local_larray,
                ]
            )
    V_new = factories.array(V_local_larray, is_split=A.split, comm=A.comm, device=A.device)
    V.balance_()
    V_new.balance_()
    V = V @ V_new

    if A.comm.rank == 0 and not silent:
        print(
            "\t" * depth
            + f"At depth {depth}: solutions of two independent problems of size {k} and {n - k} have been merged successfully."
        )

    return Lambda, V


def eigh(
    A: DNDarray,
    r_max_zolopd: int = 8,
    silent: bool = True,
) -> Tuple[DNDarray, DNDarray]:
    """
    Computes the symmetric eigenvalue decomposition of a symmetric n x n - matrix A, provided as a DNDarray.

    The function returns DNDarrays Lambda (shape (n,) with split = 0) and V (shape (n,n)) such that
    A = V @ diag(Lambda) @ V^T, where Lambda contains the eigenvalues of A and V is an orthonormal matrix
    containing the corresponding eigenvectors as columns.

    Parameters
    ----------
    A : DNDarray
        The input matrix. Must be symmetric.
    r_max_zolopd : int, optional
        This is a hyperparameter for the computation of the polar decomposition via :func:`heat.linalg.polar` which is
        applied multiple times in this function. See the documentation of :func:`heat.linalg.polar` for more details on its
        meaning and the respective default value.
    silent : bool, optional
        If True (default), suppresses output messages; otherwise, some information on the recursion is printed to the console.

    Notes
    -----
    Unlike the :func:`torch.linalg.eigh` function, the eigenvalues are returned in descending order.
    Note that no check of symmetry is performed on the input matrix A; thus, applying this function to a non-symmetric matrix may
    result in unpredictable behaviour without a specific error message pointing to this issue.

    The algorithm used for the computation of the symmetric eigenvalue decomposition is based on the Zolotarev polar decomposition;
    see Algorithm 5.2 in:

        Nakatsukasa, Y., & Freund, R. W. (2016). Computing fundamental matrix decompositions accurately via the
        matrix sign function in two iterations: The power of Zolotarev's functions. SIAM Review, 58(3).

    See Also
    --------
    :func:`heat.linalg.polar`
    """
    sanitize_in_nd_realfloating(A, "A", [2])
    if A.shape[0] != A.shape[1]:
        raise ValueError(
            f"Input matrix must be symmetric and, consequently, square, but input shape was {A.shape[0]} x {A.shape[1]}."
        )
    if not isinstance(r_max_zolopd, int) or r_max_zolopd < 1 or r_max_zolopd > 8:
        raise ValueError(
            f"If provided, parameter r_max_zolopd must be a positive integer, but was {r_max_zolopd} of type {type(r_max_zolopd)}."
        )
    return _eigh(A, None, silent, r_max_zolopd, 0, 0)



"""
Implements polar decomposition (PD)
"""

import numpy as np
import collections
import torch
from typing import Type, Callable, Dict, Any, TypeVar, Union, Tuple

from ..communication import MPICommunication, MPI
from ..dndarray import DNDarray
from .. import factories
from .. import types
from . import matrix_norm, vector_norm, matmul, qr, solve_triangular
from .basics import _estimate_largest_singularvalue, condest
from ..indexing import where
from ..random import randn
from ..devices import Device
from ..manipulations import vstack, hstack, concatenate, diag, balance
from ..exponential import sqrt
from .. import statistics

from scipy.special import ellipj
from scipy.special import ellipkm1

__all__ = ["polar"]


def _zolopd_n_iterations(r: int, kappa: float) -> int:
    """
    Returns the number of iterations required in the Zolotarev-PD algorithm.
    See the Table 3.1 in: Nakatsukasa, Y., & Freund, R. W. (2016). Computing Fundamental Matrix Decompositions Accurately via the Matrix Sign Function in Two Iterations: The Power of Zolotarev's Functions. SIAM Review, 58(3), DOI: https://doi.org/10.1137/140990334

    Inputs are `r` and `kappa` (named as in the paper), and the output is the number of iterations.
    """
    if kappa <= 1e2:
        its = [4, 3, 2, 2, 2, 2, 2, 2]
    elif kappa <= 1e3:
        its = [3, 3, 2, 2, 2, 2, 2, 2]
    elif kappa <= 1e5:
        its = [5, 3, 3, 3, 2, 2, 2, 2]
    elif kappa <= 1e7:
        its = [5, 4, 3, 3, 3, 2, 2, 2]
    else:
        its = [6, 4, 3, 3, 3, 3, 3, 2]
    return its[r - 1]


def _compute_zolotarev_coefficients(
    r: int, ell: float, device: str, dtype: types.datatype = types.float64
) -> Tuple[DNDarray, DNDarray, DNDarray]:
    """
    Computes c=(c_i)_i defined in equation (3.4), as well as a=(a_j)_j and Mhat defined in formulas (4.2)/(4.3) of the paper Nakatsukasa, Y., & Freund, R. W. (2016). Computing the polar decomposition with applications. SIAM Review, 58(3), DOI: https://doi.org/10.1137/140990334.
    Evaluations of the respective complete elliptic integral of the first kind and the Jacobi elliptic functions are imported from SciPy.

    Inputs are `r` and `ell` (named as in the paper), as well as the Heat data type `dtype` of the output (required for reasons of consistency).
    Output is a tupe containing the vectors `a` and `c` as DNDarrays and `Mhat`.
    """
    uu = np.arange(1, 2 * r + 1) * ellipkm1(ell**2) / (2 * r + 1)
    ellipfcts = np.asarray(ellipj(uu, 1 - ell**2)[:2])
    cc = ell**2 * ellipfcts[0, :] ** 2 / ellipfcts[1, :] ** 2
    aa = np.zeros(r)
    Mhat = 1
    for j in range(1, r + 1):
        p1 = 1
        p2 = 1
        for k in range(1, r + 1):
            p1 *= cc[2 * j - 2] - cc[2 * k - 1]
            if k != j:
                p2 *= cc[2 * j - 2] - cc[2 * k - 2]
        aa[j - 1] = -p1 / p2
        Mhat *= (1 + cc[2 * j - 2]) / (1 + cc[2 * j - 1])
    return (
        factories.array(cc, dtype=dtype, split=None, device=device),
        factories.array(aa, dtype=dtype, split=None, device=device),
        factories.array(Mhat, dtype=dtype, split=None, device=device),
    )


def _in_place_qr_with_q_only(A: DNDarray, procs_to_merge: int = 2) -> None:
    r"""
    Input A and procs_to_merge are as in heat.linalg.qr; difference it that this routine modified A in place and replaces it with Q.
    """
    if not A.is_distributed() or A.split < A.ndim - 2:
        # handle the case of a single process or split=None: just PyTorch QR
        # difference to heat.linalg.qr: we only return Q and put it directly in place of A
        A.larray, R = torch.linalg.qr(A.larray, mode="reduced")
        del R

    elif A.split == A.ndim - 1:
        # handle the case that A is split along the columns
        # unlike in heat.linalg.qr, we know by assumption of Zolo-PD that A has at least as many rows as columns

        nprocs = A.comm.size
        with torch.no_grad():
            for i in range(nprocs):
                # this loop goes through all the column-blocks (i.e. local arrays) of the matrix
                # this corresponds to the loop over all columns in classical Gram-Schmidt
                A_lshapes = A.lshape_map
                if i < nprocs - 1:
                    if A.comm.rank > i:
                        Q_buf = torch.zeros(
                            tuple(A_lshapes[i, :]),
                            dtype=A.larray.dtype,
                            device=A.device.torch_device,
                        )
                    color = 0 if A.comm.rank < i else 1
                    sub_comm = A.comm.Split(color, A.comm.rank)

                if A.comm.rank == i:
                    # orthogonalize the current block of columns by utilizing PyTorch QR
                    Q, R = torch.linalg.qr(A.larray, mode="reduced")
                    del R
                    A.larray[...] = Q
                    del Q
                    if i < nprocs - 1:
                        Q_buf = A.larray

                if i < nprocs - 1 and A.comm.rank >= i:
                    sub_comm.Bcast(Q_buf, root=0)

                if A.comm.rank > i:
                    # subtract the contribution of the current block of columns from the remaining columns
                    R_loc = torch.transpose(Q_buf, -2, -1) @ A.larray
                    A.larray -= Q_buf @ R_loc
                    del R_loc, Q_buf

    else:
        A, r = qr(A)
        del r


def polar(
    A: DNDarray,
    r: int = None,
    calcH: bool = True,
    condition_estimate: float = 1.0e16,
    silent: bool = True,
    r_max: int = 8,
) -> Tuple[DNDarray, DNDarray]:
    """
    Computes the so-called polar decomposition of the input 2D DNDarray ``A``, i.e., it returns the orthogonal matrix ``U`` and the symmetric, positive definite
    matrix ``H`` such that ``A = U @ H``.

    Input
    -----
    A : ht.DNDarray,
        The input matrix for which the polar decomposition is computed;
        must be two-dimensional, of data type float32 or float64, and must have at least as many rows as columns.
    r : int, optional, default: None
        The parameter r used in the Zolotarev-PD algorithm; if provided, must be an integer between 1 and 8 that divides the number of MPI processes.
        Higher values of r lead to faster convergence, but memory consumption is proportional to r.
        If not provided, the largest 1 <= r <= r_max that divides the number of MPI processes is chosen.
    calcH : bool, optional, default: True
        If True, the function returns the symmetric, positive definite matrix H. If False, only the orthogonal matrix U is returned.
    condition_estimate : float, optional, default: 1.e16.
        This argument allows to provide an estimate for the condition number of the input matrix ``A``, if such estimate is already known.
        If a positive number greater than 1., this value is used as an estimate for the condition number of A.
        If smaller or equal than 1., the condition number is estimated internally.
        The default value of 1.e16 is the worst case scenario considered in [1].
    silent : bool, optional, default: True
        If True, the function does not print any output. If False, some information is printed during the computation.
    r_max : int, optional, default: 8
        See the description of r for the meaning; r_max is only taken into account if r is not provided.


    Notes
    -----
    The implementation follows Algorithm 5.1 in Reference [1]; however, instead of switching from QR to Cholesky decomposition depending on the condition number,
    we stick to QR decomposition in all iterations.

    References
    ----------
    [1] Nakatsukasa, Y., & Freund, R. W. (2016). Computing Fundamental Matrix Decompositions Accurately via the Matrix Sign Function in Two Iterations: The Power of Zolotarev's Functions. SIAM Review, 58(3), DOI: https://doi.org/10.1137/140990334.
    """
    # check whether input is DNDarray of correct shape
    if not isinstance(A, DNDarray):
        raise TypeError(f"Input ``A`` needs to be a DNDarray but is {type(A)}.")
    if not A.ndim == 2:
        raise ValueError(f"Input ``A`` needs to be a 2D DNDarray, but its dimension is {A.ndim}.")
    if A.shape[0] < A.shape[1]:
        raise ValueError(
            f"Input ``A`` must have at least as many rows as columns, but has shape {A.shape}."
        )

    # check if A is a real floating point matrix and choose tolerances tol accordingly
    if A.dtype == types.float32:
        tol = 1.19e-7
    elif A.dtype == types.float64:
        tol = 2.22e-16
    else:
        raise TypeError(
            f"Input ``A`` must be of data type float32 or float64 but has data type {A.dtype}"
        )

    # check if input for r is reasonable
    if r is not None:
        if not isinstance(r, int) or r < 1 or r > 8:
            raise ValueError(
                f"If specified, input ``r`` must be an integer between 1 and 8, but is {r} of data type {type(r)}."
            )
        if A.is_distributed() and (A.comm.size % r != 0 or A.comm.size == r):
            raise ValueError(
                f"If specified, input ``r`` must be a non-trivial divisor of the number MPI processes , but r={r} and A.comm.size={A.comm.size}."
            )
    else:
        if not isinstance(r_max, int) or r_max < 1 or r_max > 8:
            raise ValueError(
                f"If specified, input ``r_max`` must be an integer between 1 and 8, but is {r_max} of data type {type(r_max)}."
            )
        for i in range(r_max, 0, -1):
            if A.comm.size % i == 0 and A.comm.size // i > 1:
                r = i
                break
        if not silent:
            if A.comm.rank == 0:
                print(f"Automatically chosen r={r} (r_max = {r_max}, {A.comm.size} processes).")

    # check if input for condition_estimate is reasonable
    if not isinstance(condition_estimate, float):
        raise TypeError(
            f"If specified, input ``condition_estimate`` must be a float but is {type(condition_estimate)}."
        )

    # early out for the non-distributed case
    if not A.is_distributed():
        U, s, vh = torch.linalg.svd(A.larray, full_matrices=False)
        U @= vh
        H = vh.T @ torch.diag(s) @ vh
        if calcH:
            return factories.array(U, is_split=None, comm=A.comm), factories.array(
                H, is_split=None, comm=A.comm
            )
        else:
            return factories.array(U, is_split=None, comm=A.comm)

    alpha = _estimate_largest_singularvalue(A).item()

    if condition_estimate <= 1.0:
        kappa = condest(A).item()
    else:
        kappa = condition_estimate

    if A.comm.rank == 0 and not silent:
        print(
            f"Condition number estimate: {kappa:2.2e} / Estimate for largest singular value: {alpha:2.2e}."
        )

    # each of these communicators has size r, along these communicators we parallelize the r many QR decompositions that are performed in parallel
    horizontal_comm = A.comm.Split(A.comm.rank // r, A.comm.rank)

    # each of these communicators has size MPI_WORLD.size / r and will carray a full copy of X for QR decomposition
    vertical_comm = A.comm.Split(A.comm.rank % r, A.comm.rank)

    # in each horizontal communicator, collect the local array of X from all processes
    local_shapes = horizontal_comm.allgather(A.lshape[A.split])
    new_local_shape = (
        (sum(local_shapes), A.shape[1]) if A.split == 0 else (A.shape[0], sum(local_shapes))
    )
    counts = tuple(local_shapes)
    displacements = tuple(np.cumsum([0] + list(local_shapes))[:-1])
    X_collected_local = torch.zeros(
        new_local_shape, dtype=A.dtype.torch_type(), device=A.device.torch_device
    )
    horizontal_comm.Allgatherv(
        A.larray, (X_collected_local, counts, displacements), recv_axis=A.split
    )

    X = factories.array(X_collected_local, is_split=A.split, comm=vertical_comm)
    X.balance_()
    X /= alpha

    # iteration counter and maximum number of iterations
    it = 0
    itmax = _zolopd_n_iterations(r, kappa)

    # parameters and coefficients, see Ref. [1] for their meaning
    ell = 1.0 / kappa
    c, a, Mhat = _compute_zolotarev_coefficients(r, ell, A.device, dtype=A.dtype)

    itmax = _zolopd_n_iterations(r, kappa)
    while it < itmax:
        it += 1
        if not silent:
            if A.comm.rank == 0:
                print(f"Starting Zolotarev-PD iteration no. {it}...")
        # remember current X for later convergence check
        X_old = X.copy()
        cId = factories.eye(X.shape[1], dtype=X.dtype, comm=X.comm, split=X.split, device=X.device)
        cId *= c[2 * horizontal_comm.rank].item() ** 0.5
        X = concatenate([X, cId], axis=0)
        del cId
        if X.split == 0:
            Q, R = qr(X)
            del R
            Q1 = Q[: A.shape[0], :].balance()
            Q2 = Q[A.shape[0] :, :].transpose().balance()
            Q1Q2 = matmul(Q1, Q2)
            del Q1, Q2
            X = X[: A.shape[0], :].balance()
            X /= r
        else:
            _in_place_qr_with_q_only(X)
            Q1 = X[: A.shape[0], :].balance()
            Q2 = X[A.shape[0] :, :].transpose().balance()
            del X
            Q1Q2 = matmul(Q1, Q2)
            del Q1, Q2
            X = X_old / r
        X += a[horizontal_comm.rank].item() / c[2 * horizontal_comm.rank].item() ** 0.5 * Q1Q2
        del Q1Q2
        X *= Mhat.item()
        # finally, sum over the horizontal communicators
        horizontal_comm.Allreduce(MPI.IN_PLACE, X.larray, op=MPI.SUM)

        # check for convergence and break if tolerance is reached
        if it > 1 and matrix_norm(X - X_old, ord="fro") / matrix_norm(X, ord="fro") <= tol ** (
            1 / (2 * r + 1)
        ):
            if not silent:
                if A.comm.rank == 0:
                    print(f"Zolotarev-PD iteration converged after {it} iterations.")
            break
        elif it < itmax:
            # if another iteration is necessary, update coefficients and parameters for next iteration
            ellold = ell
            ell = 1
            for j in range(r):
                ell *= (ellold**2 + c[2 * j + 1].item()) / (ellold**2 + c[2 * j].item())
            ell *= Mhat.item() * ellold
            if ell >= 1.0:
                ell = 1.0 - tol
            c, a, Mhat = _compute_zolotarev_coefficients(r, ell, A.device, dtype=A.dtype)
        else:
            if not silent:
                if A.comm.rank == 0:
                    print(
                        f"Zolotarev-PD iteration did not reach the convergence criterion after {itmax} iterations, which is most likely due to limited numerical accuracy and/or poor estimation of the condition number. The result may still be useful, but should be handled with care!"
                    )

    # as every process has much more data than required, we need to split the result into the parts that are actually
    counts = [
        X.lshape[X.split] // horizontal_comm.size + (r < X.lshape[X.split] % horizontal_comm.size)
        for r in range(horizontal_comm.size)
    ]
    displacements = [sum(counts[:r]) for r in range(horizontal_comm.size)]

    if A.split == 1:
        U_local = X.larray[
            :,
            displacements[horizontal_comm.rank] : displacements[horizontal_comm.rank]
            + counts[horizontal_comm.rank],
        ]
    else:
        U_local = X.larray[
            displacements[horizontal_comm.rank] : displacements[horizontal_comm.rank]
            + counts[horizontal_comm.rank],
            :,
        ]
    U = factories.array(U_local, is_split=A.split, comm=A.comm, device=A.device)
    del X
    U.balance_()

    # postprocessing: compute H if requested
    if calcH:
        H = matmul(U.T, A)
        H = 0.5 * (H + H.T.resplit(H.split))
        return U, H.resplit(A.split)
    else:
        return U



import numpy as np
import torch

import heat as ht
from heat.testing.basic_test import TestCase


class TestBasicTest(TestCase):
    def test_random_seed(self):
        self.assertTrue(ht.random.get_state()[1] == 42)

    def test_assert_array_equal(self):
        heat_array = ht.ones((self.get_size(), 10, 10), dtype=ht.int32, split=1)
        np_array = np.ones((self.get_size(), 10, 10), dtype=np.int32)
        self.assert_array_equal(heat_array, np_array)

        np_array[0, 1, 1] = 0
        with self.assertRaises(AssertionError):
            self.assert_array_equal(heat_array, np_array)

        heat_array = ht.zeros((25, 13, self.get_size(), 20), dtype=ht.float32, split=2)
        expected_array = torch.zeros(
            (25, 13, self.get_size(), 20),
            dtype=torch.float32,
            device=heat_array.device.torch_device,
        )
        self.assert_array_equal(heat_array, expected_array)

        if self.get_rank() == 0:
            data = torch.arange(self.get_size(), dtype=torch.int32, device=self.device.torch_device)
        else:
            data = torch.empty((0,), dtype=torch.int32, device=self.device.torch_device)

        ht_array = ht.array(data, is_split=0)
        np_array = np.arange(self.get_size(), dtype=np.int32)
        self.assert_array_equal(ht_array, np_array)

    def test_assert_func_equal(self):
        # Testing with random values
        shape = (5, 3, 2, 9)

        self.assert_func_equal(shape, heat_func=ht.exp, numpy_func=np.exp, low=-10, high=10)

        self.assert_func_equal(shape, heat_func=ht.exp2, numpy_func=np.exp2, low=-10, high=10)

        # np.random.randn eventually creates values < 0 which will result in math.nan.
        # Because math.nan != math.nan this would always produce an exception.
        self.assert_func_equal(
            shape, heat_func=ht.log, numpy_func=np.log, data_types=[np.int32, np.int64], low=1
        )

        with self.assertRaises(AssertionError):
            self.assert_func_equal(shape, heat_func=ht.exp, numpy_func=np.exp2, low=-10, high=10)

        with self.assertRaises(ValueError):
            self.assert_func_equal(np.ones(shape), heat_func=np.exp, numpy_func=np.exp)

        with self.assertRaises(ValueError):
            self.assert_func_equal(
                shape, heat_func=ht.exp, numpy_func=np.exp, low=-100, high=100, data_types=[object]
            )

    def test_assert_func_equal_for_tensor(self):
        array = np.ones((self.get_size(), 20), dtype=np.int8)
        ht_func = ht.any
        np_func = np.any
        self.assert_func_equal_for_tensor(array, ht_func, np_func, distributed_result=False)

        array = np.array([[1, 2, 4, 1, 3], [1, 4, 7, 5, 1]], dtype=np.int8)
        ht_func = ht.expand_dims
        np_func = np.expand_dims
        ht_args = {"axis": 1}
        np_args = {"axis": 1}
        self.assert_func_equal_for_tensor(
            array, ht_func, np_func, heat_args=ht_args, numpy_args=np_args
        )

        array = torch.randn(15, 15, device=self.device.torch_device)
        ht_func = ht.exp
        np_func = np.exp
        self.assert_func_equal_for_tensor(array, heat_func=ht_func, numpy_func=np_func)

        array = ht.ones((15, 15))
        with self.assertRaises(TypeError):
            self.assert_func_equal_for_tensor(array, heat_func=ht_func, numpy_func=np_func)

    def test_assertTrue_memory_layout(self):
        data = torch.arange(3 * 4 * 5).reshape(3, 4, 5)
        data_F = ht.array(data, order="F")
        with self.assertRaises(ValueError):
            self.assertTrue_memory_layout(data_F, order="K")

    def test_tmpdir(self):
        import os

        # test directory creation
        path, tmpdir = self.get_tmpdir()
        assert os.path.exists(path), f'No directory at {path}'
        assert os.path.isdir(path), f'No directory at {path}'
        ht.comm.Barrier()

        # test cleanup
        del tmpdir
        ht.comm.Barrier()
        assert not os.path.exists(path)



from __future__ import annotations

from ._dtypes import (
    _numeric_dtypes,
    _boolean_dtypes,
    _integer_dtypes,
    _integer_or_boolean_dtypes,
    _floating_dtypes,
    _result_type,
)
from ._array_object import Array
from ._data_type_functions import astype

import heat as ht


def abs(x: Array, /) -> Array:
    """
    Calculates the absolute value for each element ``x_i`` of the input array ``x``
    (i.e., the element-wise result has the same magnitude as the respective
    element in ``x`` but has positive sign).

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in abs")
    return Array._new(ht.abs(x._array, dtype=x.dtype))


def acos(x: Array, /) -> Array:
    """
    Calculates an approximation of the principal value of the inverse cosine,
    having domain ``[-1, +1]`` and codomain ``[+0, +π]``, for each element ``x_i``
    of the input array ``x``. Each element-wise result is expressed in radians.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in acos")
    return Array._new(ht.acos(x._array))


def acosh(x: Array, /) -> Array:
    """
    Calculates an approximation to the inverse hyperbolic cosine, having domain
    ``[+1, +infinity]`` and codomain ``[+0, +infinity]``, for each element ``x_i``
    of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent the area of a hyperbolic sector.
        Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in acosh")
    return Array._new(ht.acosh(x._array))


def add(x1: Array, x2: Array, /) -> Array:
    """
    Calculates the sum for each element ``x1_i`` of the input array ``x1`` with
    the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in add")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.add(x1._array, x2._array))


def asin(x: Array, /) -> Array:
    """
    Calculates an approximation of the principal value of the inverse sine, having
    domain ``[-1, +1]`` and codomain ``[-π/2, +π/2]`` for each element ``x_i`` of
    the input array ``x``. Each element-wise result is expressed in radians.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in asin")
    return Array._new(ht.asin(x._array))


def asinh(x: Array, /) -> Array:
    """
    Calculates an approximation to the inverse hyperbolic sine, having domain
    ``[-infinity, +infinity]`` and codomain ``[-infinity, +infinity]``, for each
    element ``x_i`` in the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent the area of a hyperbolic sector.
        Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in asinh")
    return Array._new(ht.asinh(x._array))


def atan(x: Array, /) -> Array:
    """
    Calculates an implementation-dependent approximation of the principal value of
    the inverse tangent, having domain ``[-infinity, +infinity]`` and codomain
    ``[-π/2, +π/2]``, for each element ``x_i`` of the input array ``x``.
    Each element-wise result is expressed in radians.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in atan")
    return Array._new(ht.atan(x._array))


def atan2(x1: Array, x2: Array, /) -> Array:
    """
    Calculates an approximation of the inverse tangent of the quotient ``x1/x2``,
    having domain ``[-infinity, +infinity] x [-infinity, +infinity]`` (where the
    ``x`` notation denotes the set of ordered pairs of elements ``(x1_i, x2_i)``)
    and codomain ``[-π, +π]``, for each pair of elements ``(x1_i, x2_i)`` of the
    input arrays ``x1`` and ``x2``, respectively. Each element-wise result is
    expressed in radians.

    The mathematical signs of ``x1_i`` and ``x2_i`` determine the quadrant of each
    element-wise result. The quadrant (i.e., branch) is chosen such that each
    element-wise result is the signed angle in radians between the ray ending at the
    origin and passing through the point ``(1,0)`` and the ray ending at the origin
    and passing through the point ``(x2_i, x1_i)``.

    Parameters
    ----------
    x1 : Array
        Input array corresponding to the y-coordinates. Must have a floating-point
        data type.
    x2 : Array
        Input array corresponding to the x-coordinates. Must be compatible with ``x1``
        and have a floating-point data type.
    """
    if x1.dtype not in _floating_dtypes or x2.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in atan2")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.atan2(x1._array, x2._array))


def atanh(x: Array, /) -> Array:
    """
    Calculates an approximation to the inverse hyperbolic tangent, having domain
    ``[-1, +1]`` and codomain ``[-infinity, +infinity]``, for each element ``x_i``
    of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent the area of a hyperbolic sector.
        Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in atanh")
    return Array._new(ht.atanh(x._array))


def bitwise_and(x1: Array, x2: Array, /) -> Array:
    """
    Computes the bitwise AND of the underlying binary representation of each
    element ``x1_i`` of the input array ``x1`` with the respective element
    ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have an integer or boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have an integer
        or boolean data type.
    """
    if x1.dtype not in _integer_or_boolean_dtypes or x2.dtype not in _integer_or_boolean_dtypes:
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_and")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.bitwise_and(x1._array, x2._array))


def bitwise_left_shift(x1: Array, x2: Array, /) -> Array:
    """
    Shifts the bits of each element ``x1_i`` of the input array ``x1`` to the
    left by appending ``x2_i`` (i.e., the respective element in the input array
    ``x2``) zeros to the right of ``x1_i``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have an integer data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have an integer
        data type. Each element must be greater than or equal to ``0``.
    """
    if x1.dtype not in _integer_dtypes or x2.dtype not in _integer_dtypes:
        raise TypeError("Only integer dtypes are allowed in bitwise_left_shift")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    if ht.any(x2._array < 0):
        raise ValueError("bitwise_left_shift(x1, x2) is only defined for x2 >= 0")
    return Array._new(ht.left_shift(x1._array, x2._array))


def bitwise_invert(x: Array, /) -> Array:
    """
    Inverts (flips) each bit for each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have an integer or boolean data type.
    """
    if x.dtype not in _integer_or_boolean_dtypes:
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_invert")
    return Array._new(ht.invert(x._array))


def bitwise_or(x1: Array, x2: Array, /) -> Array:
    """
    Computes the bitwise OR of the underlying binary representation of each
    element ``x1_i`` of the input array ``x1`` with the respective element
    ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have an integer or boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have an integer
        or boolean data type.
    """
    if x1.dtype not in _integer_or_boolean_dtypes or x2.dtype not in _integer_or_boolean_dtypes:
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_or")
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.bitwise_or(x1._array, x2._array))


def bitwise_right_shift(x1: Array, x2: Array, /) -> Array:
    """
    Shifts the bits of each element ``x1_i`` of the input array ``x1`` to the
    right according to the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have an integer data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have an integer
        data type. Each element must be greater than or equal to ``0``.
    """
    if x1.dtype not in _integer_dtypes or x2.dtype not in _integer_dtypes:
        raise TypeError("Only integer dtypes are allowed in bitwise_right_shift")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    # Note: bitwise_right_shift is only defined for x2 nonnegative.
    if ht.any(x2._array < 0):
        raise ValueError("bitwise_right_shift(x1, x2) is only defined for x2 >= 0")
    return Array._new(ht.right_shift(x1._array, x2._array))


def bitwise_xor(x1: Array, x2: Array, /) -> Array:
    """
    Computes the bitwise XOR of the underlying binary representation of each element
    ``x1_i`` of the input array ``x1`` with the respective element ``x2_i`` of the
    input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have an integer or boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have an integer
        or boolean data type.
    """
    if x1.dtype not in _integer_or_boolean_dtypes or x2.dtype not in _integer_or_boolean_dtypes:
        raise TypeError("Only integer or boolean dtypes are allowed in bitwise_xor")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.bitwise_xor(x1._array, x2._array))


def ceil(x: Array, /) -> Array:
    """
    Rounds each element ``x_i`` of the input array ``x`` to the smallest (i.e., closest
    to ``-infinity``) integer-valued number that is not less than ``x_i``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in ceil")
    if x.dtype in _integer_dtypes:
        return x
    return Array._new(ht.ceil(x._array))


def cos(x: Array, /) -> Array:
    """
    Calculates an approximation to the cosine, having domain ``(-infinity, +infinity)``
    and codomain ``[-1, +1]``, for each element ``x_i`` of the input array ``x``.
    Each element ``x_i`` is assumed to be expressed in radians.

    Parameters
    ----------
    x : Array
        Input array whose elements are each expressed in radians. Must have a
        floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in cos")
    return Array._new(ht.cos(x._array))


def cosh(x: Array, /) -> Array:
    """
    Calculates an approximation to the hyperbolic cosine, having domain
    ``[-infinity, +infinity]`` and codomain ``[-infinity, +infinity]``, for each
    element ``x_i`` in the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent a hyperbolic angle. Must have
        a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in cosh")
    return Array._new(ht.cosh(x._array))


def divide(x1: Array, x2: Array, /) -> Array:
    """
    Calculates the division for each element ``x1_i`` of the input array ``x1``
    with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        Dividend input array. Must have a numeric data type.
    x2 : Array
        Divisor input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _floating_dtypes or x2.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in divide")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.divide(x1._array, x2._array))


def equal(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i == x2_i`` for each element ``x1_i`` of
    the input array ``x1`` with the respective element ``x2_i`` of the input
    array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array.
    x2 : Array
        Second input array. Must be compatible with ``x1``.
    """
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.eq(x1._array, x2._array))


def exp(x: Array, /) -> Array:
    """
    Calculates an approximation to the exponential function, having domain
    ``[-infinity, +infinity]`` and codomain ``[+0, +infinity]``, for each element
    ``x_i`` of the input array ``x`` (``e`` raised to the power of ``x_i``, where
    ``e`` is the base of the natural logarithm).

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in exp")
    return Array._new(ht.exp(x._array))


def expm1(x: Array, /) -> Array:
    """
    Calculates an approximation to ``exp(x)-1``, having domain
    ``[-infinity, +infinity]`` and codomain ``[-1, +infinity]``, for each element
    ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in expm1")
    return Array._new(ht.expm1(x._array))


def floor(x: Array, /) -> Array:
    """
    Rounds each element ``x_i`` of the input array ``x`` to the greatest
    (i.e., closest to ``+infinity``) integer-valued number that is not greater
    than ``x_i``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in floor")
    if x.dtype in _integer_dtypes:
        return x
    return Array._new(ht.floor(x._array))


def floor_divide(x1: Array, x2: Array, /) -> Array:
    """
    Rounds the result of dividing each element ``x1_i`` of the input array ``x1``
    by the respective element ``x2_i`` of the input array ``x2`` to the greatest
    (i.e., closest to ``+infinity``) integer-value number that is not greater than
    the division result.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in floor_divide")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.floor_divide(x1._array, x2._array))


def greater(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i > x2_i`` for each element ``x1_i`` of
    the input array ``x1`` with the respective element ``x2_i`` of the input
    array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in greater")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.greater(x1._array, x2._array))


def greater_equal(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i >= x2_i`` for each element ``x1_i`` of
    the input array ``x1`` with the respective element ``x2_i`` of the input
    array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in greater_equal")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.greater_equal(x1._array, x2._array))


def isfinite(x: Array, /) -> Array:
    """
    Tests each element ``x_i`` of the input array ``x`` to determine if finite
    (i.e., not ``NaN`` and not equal to positive or negative infinity).

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isfinite")
    return Array._new(ht.isfinite(x._array))


def isinf(x: Array, /) -> Array:
    """
    Tests each element ``x_i`` of the input array ``x`` to determine if equal
    to positive or negative infinity.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isinf")
    return Array._new(ht.isinf(x._array))


def isnan(x: Array, /) -> Array:
    """
    Tests each element ``x_i`` of the input array ``x`` to determine whether
    the element is ``NaN``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in isnan")
    return Array._new(ht.isnan(x._array))


def less(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i < x2_i`` for each element ``x1_i`` of
    the input array ``x1`` with the respective element ``x2_i`` of the input
    array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in less")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.less(x1._array, x2._array))


def less_equal(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i <= x2_i`` for each element ``x1_i``
    of the input array ``x1`` with the respective element ``x2_i`` of the
    input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in less_equal")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.less_equal(x1._array, x2._array))


def log(x: Array, /) -> Array:
    """
    Calculates an  approximation to the natural (base ``e``) logarithm, having domain
    ``[0, +infinity]`` and codomain ``[-infinity, +infinity]``, for each element
    ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log")
    return Array._new(ht.log(x._array))


def log1p(x: Array, /) -> Array:
    """
    Calculates an approximation to ``log(1+x)``, where ``log`` refers to the natural
    (base ``e``) logarithm, having domain ``[-1, +infinity]`` and codomain
    ``[-infinity, +infinity]``, for each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log1p")
    return Array._new(ht.log1p(x._array))


def log2(x: Array, /) -> Array:
    """
    Calculates an approximation to the base ``2`` logarithm, having domain
    ``[0, +infinity]`` and codomain ``[-infinity, +infinity]``, for each element
    ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log2")
    return Array._new(ht.log2(x._array))


def log10(x: Array, /) -> Array:
    """
    Calculates an approximation to the base ``10`` logarithm, having domain
    ``[0, +infinity]`` and codomain ``[-infinity, +infinity]``, for each element
    ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in log10")
    return Array._new(ht.log10(x._array))


def logaddexp(x1: Array, x2: Array) -> Array:
    """
    Calculates the logarithm of the sum of exponentiations ``log(exp(x1) + exp(x2))``
    for each element ``x1_i`` of the input array ``x1`` with the respective element
    ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a floating-point data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a floating-point
        data type.
    """
    if x1.dtype not in _floating_dtypes or x2.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in logaddexp")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.logaddexp(x1._array, x2._array))


def logical_and(x1: Array, x2: Array, /) -> Array:
    """
    Computes the logical AND for each element ``x1_i`` of the input array ``x1``
    with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a boolean
        data type.
    """
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_and")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.logical_and(x1._array, x2._array))


def logical_not(x: Array, /) -> Array:
    """
    Computes the logical NOT for each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a boolean data type.
    """
    if x.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_not")
    return Array._new(ht.logical_not(x._array))


def logical_or(x1: Array, x2: Array, /) -> Array:
    """
    Computes the logical OR for each element ``x1_i`` of the input array ``x1``
    with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a boolean
        data type.
    """
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_or")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.logical_or(x1._array, x2._array))


def logical_xor(x1: Array, x2: Array, /) -> Array:
    """
    Computes the logical XOR for each element ``x1_i`` of the input array ``x1``
    with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a boolean data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a boolean
        data type.
    """
    if x1.dtype not in _boolean_dtypes or x2.dtype not in _boolean_dtypes:
        raise TypeError("Only boolean dtypes are allowed in logical_xor")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.logical_xor(x1._array, x2._array))


def multiply(x1: Array, x2: Array, /) -> Array:
    """
    Calculates the product for each element ``x1_i`` of the input array ``x1``
    with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in multiply")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.multiply(x1._array, x2._array))


def negative(x: Array, /) -> Array:
    """
    Computes the numerical negative of each element ``x_i``
    (i.e., ``y_i = -x_i``) of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in negative")
    return Array._new(ht.negative(x._array))


def not_equal(x1: Array, x2: Array, /) -> Array:
    """
    Computes the truth value of ``x1_i != x2_i`` for each element ``x1_i``
    of the input array ``x1`` with the respective element ``x2_i`` of the
    input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array.
    x2 : Array
        Second input array. Must be compatible with ``x1``.
    """
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.not_equal(x1._array, x2._array))


def positive(x: Array, /) -> Array:
    """
    Computes the numerical positive of each element ``x_i``
    (i.e., ``y_i = +x_i``) of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in positive")
    return Array._new(ht.positive(x._array))


def pow(x1: Array, x2: Array, /) -> Array:
    """
    Calculates an approximation of exponentiation by raising each element
    ``x1_i`` (the base) of the input array ``x1`` to the power of
    ``x2_i`` (the exponent), where ``x2_i`` is the corresponding element of
    the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array whose elements correspond to the exponentiation base.
        Must have a numeric data type.
    x2 : Array
        Second input array whose elements correspond to the exponentiation exponent.
        Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in pow")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.pow(x1._array, x2._array))


def remainder(x1: Array, x2: Array, /) -> Array:
    """
    Returns the remainder of division for each element ``x1_i`` of the input
    array ``x1`` and the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        Dividend input array. Must have a numeric data type.
    x2 : Array
        Divisor input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in remainder")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.remainder(x1._array, x2._array))


def round(x: Array, /) -> Array:
    """
    Rounds each element ``x_i`` of the input array ``x`` to the nearest
    integer-valued number.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in round")
    elif x.dtype in _integer_dtypes:
        return x
    return Array._new(ht.round(x._array))


def sign(x: Array, /) -> Array:
    """
    Returns an indication of the sign of a number for each element ``x_i`` of
    the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in sign")
    return astype(Array._new(ht.sign(x._array)), x.dtype)


def sin(x: Array, /) -> Array:
    """
    Calculates an approximation to the sine, having domain ``(-infinity, +infinity)``
    and codomain ``[-1, +1]``, for each element ``x_i`` of the input array ``x``.
    Each element ``x_i`` is assumed to be expressed in radians.

    Parameters
    ----------
    x : Array
        Input array whose elements are each expressed in radians. Must have a
        floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sin")
    return Array._new(ht.sin(x._array))


def sinh(x: Array, /) -> Array:
    """
    Calculates an approximation to the hyperbolic sine, having domain
    ``[-infinity, +infinity]`` and codomain ``[-infinity, +infinity]``, for
    each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent a hyperbolic angle. Must have
        a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sinh")
    return Array._new(ht.sinh(x._array))


def square(x: Array, /) -> Array:
    """
    Squares ``(x_i * x_i)`` each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in square")
    return astype(Array._new(ht.square(x._array)), x.dtype)


def sqrt(x: Array, /) -> Array:
    """
    Calculates the square root, having domain ``[0, +infinity]`` and codomain
    ``[0, +infinity]``, for each element ``x_i`` of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array. Must have a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in sqrt")
    return Array._new(ht.sqrt(x._array))


def subtract(x1: Array, x2: Array, /) -> Array:
    """
    Calculates the difference for each element ``x1_i`` of the input array
    ``x1`` with the respective element ``x2_i`` of the input array ``x2``.

    Parameters
    ----------
    x1 : Array
        First input array. Must have a numeric data type.
    x2 : Array
        Second input array. Must be compatible with ``x1`` and have a numeric
        data type.
    """
    if x1.dtype not in _numeric_dtypes or x2.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in subtract")
    # Call result type here just to raise on disallowed type combinations
    _result_type(x1.dtype, x2.dtype)
    x1, x2 = Array._normalize_two_args(x1, x2)
    return Array._new(ht.subtract(x1._array, x2._array))


def tan(x: Array, /) -> Array:
    """
    Calculates an approximation to the tangent, having domain ``(-infinity, +infinity)``
    and codomain ``(-infinity, +infinity)``, for each element ``x_i`` of the
    input array ``x``. Each element ``x_i`` is assumed to be expressed in radians.

    Parameters
    ----------
    x : Array
        Input array whose elements are each expressed in radians. Must have a
        floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in tan")
    return Array._new(ht.tan(x._array))


def tanh(x: Array, /) -> Array:
    """
    Calculates an approximation to the hyperbolic tangent, having domain
    ``[-infinity, +infinity]`` and codomain ``[-1, +1]``, for each element ``x_i``
    of the input array ``x``.

    Parameters
    ----------
    x : Array
        Input array whose elements each represent a hyperbolic angle. Must have
        a floating-point data type.
    """
    if x.dtype not in _floating_dtypes:
        raise TypeError("Only floating-point dtypes are allowed in tanh")
    return Array._new(ht.tanh(x._array))


def trunc(x: Array, /) -> Array:
    """
    Rounds each element ``x_i`` of the input array ``x`` to the integer-valued
    number that is closest to but no greater than ``x_i``.

    Parameters
    ----------
    x : Array
        Input array. Must have a numeric data type.
    """
    if x.dtype not in _numeric_dtypes:
        raise TypeError("Only numeric dtypes are allowed in trunc")
    if x.dtype in _integer_dtypes:
        return x
    return Array._new(ht.trunc(x._array))



"""Provides a collection of signal-processing operations"""

import torch
import numpy as np

from .communication import MPI
from .dndarray import DNDarray
from .types import promote_types, float32, float64
from .manipulations import pad, flip
from .factories import array, zeros, arange
import torch.nn.functional as fc

__all__ = ["convolve"]


def convolve(a: DNDarray, v: DNDarray, mode: str = "full", stride: int = 1) -> DNDarray:
    """
    Returns the discrete, linear convolution of two one-dimensional `DNDarray`s or scalars.
    Unlike `numpy.signal.convolve`, if ``a`` and/or ``v`` have more than one dimension, batch-convolution along the last dimension will be attempted. See `Examples` below.

    Parameters
    ----------
    a : DNDarray or scalar
        One- or N-dimensional signal ``DNDarray`` of shape (..., N), or scalar. If ``a`` has more than one dimension, it will be treated as a batch of 1D signals.
        Distribution along the batch dimension is required for distributed batch processing. See the examples for details.
    v : DNDarray or scalar
        One- or N-dimensional filter weight `DNDarray` of shape (..., M), or scalar. If ``v`` has more than one dimension, it will be treated as a batch of 1D filter weights.
        The batch dimension(s) of ``v`` must match the batch dimension(s) of ``a``.
    mode : str
        Can be 'full', 'valid', or 'same'. Default is 'full'.
        'full':
          Returns the convolution at
          each point of overlap, with a length of '(N+M-2)//stride+1'. At
          the end-points of the convolution, the signals do not overlap
          completely, and boundary effects may be seen.
        'same':
          Mode 'same' returns output  of length 'N'. Boundary
          effects are still visible. This mode is not supported for
          even-sized filter weights
        'valid':
          Mode 'valid' returns output of length '(N-M)//stride+1'. The
          convolution product is only given for points where the signals
          overlap completely. Values outside the signal boundary have no
          effect.
    stride : int
        Stride of the convolution. Must be a positive integer. Default is 1.
        Stride must be 1 for mode 'same'.

    Examples
    --------
    Note how the convolution operator flips the second array
    before "sliding" the two across one another:

    >>> a = ht.ones(5)
    >>> v = ht.arange(3).astype(ht.float)
    >>> ht.convolve(a, v, mode="full")
    DNDarray([0., 1., 3., 3., 3., 3., 2.])
    >>> ht.convolve(a, v, mode="same")
    DNDarray([1., 3., 3., 3., 3.])
    >>> ht.convolve(a, v, mode="valid")
    DNDarray([3., 3., 3.])
    >>> ht.convolve(a, v, stride=2)
    DNDarray([0., 3., 3., 2.])
    >>> ht.convolve(a, v, mode="valid", stride=2)
    DNDarray([3., 3.])

    >>> a = ht.ones(10, split=0)
    >>> v = ht.arange(3, split=0).astype(ht.float)
    >>> ht.convolve(a, v, mode="valid")
    DNDarray([3., 3., 3., 3., 3., 3., 3., 3.])

    [0/3] DNDarray([3., 3., 3.])
    [1/3] DNDarray([3., 3., 3.])
    [2/3] DNDarray([3., 3.])

    >>> a = ht.ones(10, split=0)
    >>> v = ht.arange(3, split=0)
    >>> ht.convolve(a, v)
    DNDarray([0., 1., 3., 3., 3., 3., 3., 3., 3., 3., 3., 2.], dtype=ht.float32, device=cpu:0, split=0)

    [0/3] DNDarray([0., 1., 3., 3.])
    [1/3] DNDarray([3., 3., 3., 3.])
    [2/3] DNDarray([3., 3., 3., 2.])

    >>> a = ht.arange(50, dtype=ht.float64, split=0)
    >>> a = a.reshape(10, 5)  # 10 signals of length 5
    >>> v = ht.arange(3)
    >>> ht.convolve(a, v)  # batch processing: 10 signals convolved with filter v
    DNDarray([[  0.,   0.,   1.,   4.,   7.,  10.,   8.],
          [  0.,   5.,  16.,  19.,  22.,  25.,  18.],
          [  0.,  10.,  31.,  34.,  37.,  40.,  28.],
          [  0.,  15.,  46.,  49.,  52.,  55.,  38.],
          [  0.,  20.,  61.,  64.,  67.,  70.,  48.],
          [  0.,  25.,  76.,  79.,  82.,  85.,  58.],
          [  0.,  30.,  91.,  94.,  97., 100.,  68.],
          [  0.,  35., 106., 109., 112., 115.,  78.],
          [  0.,  40., 121., 124., 127., 130.,  88.],
          [  0.,  45., 136., 139., 142., 145.,  98.]], dtype=ht.float64, device=cpu:0, split=0)

    >>> v = ht.random.randint(0, 3, (10, 3), split=0)  # 10 filters of length 3
    >>> ht.convolve(a, v)  # batch processing: 10 signals convolved with 10 filters
    DNDarray([[  0.,   0.,   2.,   4.,   6.,   8.,   0.],
            [  5.,   6.,   7.,   8.,   9.,   0.,   0.],
            [ 20.,  42.,  56.,  61.,  66.,  41.,  14.],
            [  0.,  15.,  16.,  17.,  18.,  19.,   0.],
            [ 20.,  61.,  64.,  67.,  70.,  48.,   0.],
            [ 50.,  52., 104., 108., 112.,  56.,  58.],
            [  0.,  30.,  61.,  63.,  65.,  67.,  34.],
            [ 35., 106., 109., 112., 115.,  78.,   0.],
            [  0.,  40.,  81.,  83.,  85.,  87.,  44.],
            [  0.,   0.,  45.,  46.,  47.,  48.,  49.]], dtype=ht.float64, device=cpu:0, split=0)
    """
    if np.isscalar(a):
        a = array([a])
    if np.isscalar(v):
        v = array([v])
    if not isinstance(a, DNDarray):
        try:
            a = array(a)
        except TypeError:
            raise TypeError(f"non-supported type for signal: {type(a)}")
    if not isinstance(v, DNDarray):
        try:
            v = array(v)
        except TypeError:
            raise TypeError(f"non-supported type for filter: {type(v)}")
    promoted_type = promote_types(a.dtype, v.dtype)
    if a.larray.is_mps and promoted_type == float64:
        # cannot cast to float64 on MPS
        promoted_type = float32

    a = a.astype(promoted_type)
    v = v.astype(promoted_type)

    # check if the filter is longer than the signal and swap them if necessary
    if v.shape[-1] > a.shape[-1]:
        a, v = v, a

    # assess whether to perform batch processing, default is False (no batch processing)
    batch_processing = False
    if a.ndim > 1:
        # batch processing requires 1D filter OR matching batch dimensions for signal and filter
        batch_dims = a.shape[:-1]
        # verify that the filter shape is consistent with the signal
        if v.ndim > 1:
            if v.shape[:-1] != batch_dims:
                raise ValueError(
                    f"Batch dimensions of signal and filter must match. Signal: {a.shape}, Filter: {v.shape}"
                )
        if a.is_distributed():
            if a.split == a.ndim - 1:
                raise ValueError(
                    "Please distribute the signal along the batch dimension, not the signal dimension. For in-place redistribution use the `DNDarray.resplit_()` method with `axis=0`"
                )
            if v.is_distributed():
                if v.ndim == 1:
                    # gather filter to all ranks
                    v.resplit_(axis=None)
                else:
                    v.resplit_(axis=a.split)
        batch_processing = True

    if not batch_processing and v.ndim > 1:
        raise ValueError(
            f"1-D convolution only supported for 1-dimensional signal and kernel. Signal: {a.shape}, Filter: {v.shape}"
        )

    # check mode and stride for value errors
    if stride < 1:
        raise ValueError("Stride must be at positive integer")
    if stride > 1 and mode == "same":
        raise ValueError("Stride must be 1 for mode 'same'")

    if mode == "same" and v.shape[-1] % 2 == 0:
        raise ValueError("Mode 'same' cannot be used with even-sized kernel")
    if not v.is_balanced():
        raise ValueError("Only balanced kernel weights are allowed")

    # calculate pad size according to mode
    if mode == "full":
        pad_size = v.shape[-1] - 1
    elif mode == "same":
        pad_size = v.shape[-1] // 2
    elif mode == "valid":
        pad_size = 0
    else:
        raise ValueError(f"Supported modes are 'full', 'valid', 'same', got {mode}")

    gshape = (a.shape[-1] + 2 * pad_size - v.shape[-1]) // stride + 1

    if v.is_distributed() and stride > 1:
        gshape_stride_1 = a.shape[-1] + 2 * pad_size - v.shape[-1] + 1

    if batch_processing:
        # all operations are local torch operations, only the last dimension is convolved
        local_a = a.larray
        local_v = v.larray

        # flip filter for convolution, as Pytorch conv1d computes correlations
        local_v = torch.flip(local_v, [-1])
        local_batch_dims = tuple(local_a.shape[:-1])

        # reshape signal and filter to 3D for Pytorch conv1d function
        # see https://pytorch.org/docs/stable/generated/torch.nn.functional.conv1d.html
        local_a = local_a.reshape(
            torch.prod(torch.tensor(local_batch_dims, device=local_a.device), dim=0).item(),
            local_a.shape[-1],
        )
        channels = local_a.shape[0]
        if v.ndim > 1:
            local_v = local_v.reshape(
                torch.prod(torch.tensor(local_batch_dims, device=local_v.device), dim=0).item(),
                local_v.shape[-1],
            )
            local_v = local_v.unsqueeze(1)
        else:
            local_v = local_v.unsqueeze(0).unsqueeze(0).expand(local_a.shape[0], 1, -1)
        # add batch dimension to signal
        local_a = local_a.unsqueeze(0)

        # cast to single-precision float if on GPU
        if local_a.is_cuda:
            float_type = torch.promote_types(local_a.dtype, torch.float32)
            local_a = local_a.to(float_type)
            local_v = local_v.to(float_type)

        # apply torch convolution operator if local signal isn't empty
        if torch.prod(torch.tensor(local_a.shape, device=local_a.device)) > 0:
            local_convolved = fc.conv1d(
                local_a, local_v, padding=pad_size, groups=channels, stride=stride
            )
        else:
            empty_shape = tuple(local_a.shape[:-1] + (gshape,))
            local_convolved = torch.empty(empty_shape, dtype=local_a.dtype, device=local_a.device)

        # unpack 3D result into original shape
        local_convolved = local_convolved.squeeze(0)
        local_convolved = local_convolved.reshape(local_batch_dims + (gshape,))

        # wrap result in DNDarray
        convolved = array(local_convolved, is_split=a.split, device=a.device, comm=a.comm)
        return convolved

    # pad signal with zeros
    a = pad(a, pad_size, "constant", 0)

    # compute halo size
    halo_size = torch.max(v.lshape_map[:, -1]).item() // 2

    if a.is_distributed():
        if (v.lshape_map[:, 0] > a.lshape_map[:, 0]).any():
            raise ValueError(
                "Local chunk of filter weight is larger than the local chunks of signal"
            )
        # fetch halos and store them in a.halo_next/a.halo_prev
        a.get_halo(halo_size)
        # apply halos to local array
        signal = a.array_with_halos

        # shift signal based on global kernel starts for any rank but first
        if stride > 1 and not v.is_distributed():
            if a.comm.rank == 0:
                local_index = 0
            else:
                local_index = torch.sum(a.lshape_map[: a.comm.rank, 0]).item() - halo_size
                local_index = local_index % stride

                if local_index != 0:
                    local_index = stride - local_index

                # even kernels can produces doubles
                if v.shape[-1] % 2 == 0 and local_index == 0:
                    local_index = stride

            signal = signal[local_index:]
    else:
        signal = a.larray

    # flip filter for convolution as Pytorch conv1d computes correlations
    v = flip(v, [0])
    if v.larray.shape != v.lshape_map[0]:
        # pads weights if input kernel is uneven
        target = torch.zeros(v.lshape_map[0][0], dtype=v.larray.dtype, device=v.larray.device)
        pad_size = v.lshape_map[0][0] - v.larray.shape[0]
        target[pad_size:] = v.larray
        weight = target
    else:
        weight = v.larray

    t_v = weight  # stores temporary weight

    # make signal and filter weight 3D for Pytorch conv1d function
    signal = signal.reshape(1, 1, signal.shape[0])
    weight = weight.reshape(1, 1, weight.shape[0])

    # cast to float if on GPU
    if signal.is_cuda:
        float_type = promote_types(signal.dtype, torch.float32).torch_type()
        signal = signal.to(float_type)
        weight = weight.to(float_type)
        t_v = t_v.to(float_type)

    if v.is_distributed():
        size = v.comm.size

        # any stride is a subset of stride 1
        if stride > 1:
            gshape = gshape_stride_1

        for r in range(size):
            rec_v = t_v.clone()
            v.comm.Bcast(rec_v, root=r)
            t_v1 = rec_v.reshape(1, 1, rec_v.shape[0])
            local_signal_filtered = fc.conv1d(signal, t_v1, stride=1)
            # unpack 3D result into 1D
            local_signal_filtered = local_signal_filtered[0, 0, :]

            if a.comm.rank != 0 and v.lshape_map[0][0] % 2 == 0:
                local_signal_filtered = local_signal_filtered[1:]

            # accumulate filtered signal on the fly
            global_signal_filtered = array(
                local_signal_filtered, is_split=0, device=a.device, comm=a.comm
            )
            if r == 0:
                # initialize signal_filtered, starting point of slice
                signal_filtered = zeros(
                    gshape, dtype=a.dtype, split=a.split, device=a.device, comm=a.comm
                )
                start_idx = 0

            # accumulate relevant slice of filtered signal
            # note, this is a binary operation between unevenly distributed dndarrays and will require communication, check out _operations.__binary_op()
            try:
                signal_filtered += global_signal_filtered[start_idx : start_idx + gshape]
            except (ValueError, TypeError):
                signal_filtered = (
                    signal_filtered + global_signal_filtered[start_idx : start_idx + gshape]
                )
            if r != size - 1:
                start_idx += v.lshape_map[r + 1][0].item()

        # any stride is a subset of arrays of stride 1
        if stride > 1:
            signal_filtered = signal_filtered[::stride]

        return signal_filtered

    else:
        # apply torch convolution operator
        if signal.shape[-1] >= weight.shape[-1]:
            signal_filtered = fc.conv1d(signal, weight, stride=stride)

            # unpack 3D result into 1D
            signal_filtered = signal_filtered[0, 0, :]
        else:
            signal_filtered = torch.tensor([], device=str(signal.device))

        # if kernel shape along split axis is even we need to get rid of duplicated values
        if a.comm.rank != 0 and v.shape[0] % 2 == 0 and stride == 1:
            signal_filtered = signal_filtered[1:]

        return DNDarray(
            signal_filtered,
            (gshape,),
            signal_filtered.dtype,
            a.split,
            a.device,
            a.comm,
            balanced=False,
        ).astype(a.dtype.torch_type())



from unittest.mock import patch
import argparse
from heat import cli
import io
import contextlib

class TestCLI:
    @patch("argparse.ArgumentParser.parse_args", return_value=argparse.Namespace(info=False))
    def test_cli_help(self, mock_parse_args):
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            cli.cli()

        print(stdout.getvalue())
        assert "usage: heat [-h] [-i]" in stdout.getvalue()

    @patch("platform.platform")
    @patch("mpi4py.MPI.Get_library_version")
    @patch("torch.cuda.is_available")
    @patch("torch.cuda.device_count")
    @patch("torch.cuda.current_device")
    @patch("torch.cuda.get_device_name")
    @patch("torch.cuda.get_device_properties")
    def test_platform_info(
        self,
        mock_get_device_properties,
        mock_get_device_name,
        mock_get_default_device,
        mock_device_count,
        mock_cuda_current_device,
        mock_mpi_lib_version,
        mock_platform,
    ):
        mock_platform.return_value = "Test Platform"
        mock_mpi_lib_version.return_value = "Test MPI Library"
        mock_cuda_current_device.return_value = True
        mock_device_count.return_value = 1
        mock_get_default_device.return_value = "cuda:0"
        mock_get_device_name.return_value = "Test Device"
        mock_get_device_properties.return_value.total_memory = 1024**4  # 1TiB

        stdout_stream = io.StringIO()
        with contextlib.redirect_stdout(stdout_stream):
            cli.plaform_info()
        stdout = stdout_stream.getvalue()
        print(stdout)
        assert "HeAT: Helmholtz Analytics Toolkit" in stdout
        assert "Platform: Test Platform" in stdout
        assert "mpi4py Version:" in stdout
        assert "MPI Library Version: Test MPI Library" in stdout
        assert "Torch Version:" in stdout
        assert "CUDA Available: True" in stdout
        assert "Device count: 1" in stdout
        assert "Default device: cuda:0" in stdout
        assert "Device name: Test Device" in stdout
        assert "Device memory: 1024.0 GiB" in stdout



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, Union

if TYPE_CHECKING:
    from ._typing import Array, Device, Dtype, NestedSequence, SupportsBufferProtocol
from ._dtypes import _all_dtypes, _floating_dtypes, default_float, default_int, bool as api_bool

import heat as ht
import torch


def arange(
    start: Union[int, float],
    /,
    stop: Optional[Union[int, float]] = None,
    step: Union[int, float] = 1,
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns evenly spaced values within the half-open interval ``[start, stop)``
    as a one-dimensional array.

    Parameters
    ----------
    start : Union[int, float]
        If ``stop`` is specified, the start of interval (inclusive); otherwise,
        the end of the interval (exclusive). If ``stop`` is not specified,
        the default starting value is ``0``.
    stop : Optional[Union[int, float]]
        The end of the interval. Default: ``None``.
    step : Union[int, float]
        the distance between two adjacent elements (``out[i+1] - out[i]``). Must
        not be ``0``; may be negative, this results in an empty array if
        ``stop >= start``. Default: ``1``.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from ``start``, ``stop`` and ``step``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    if dtype is None:
        if isinstance(start, float) or isinstance(stop, float) or isinstance(step, float):
            dtype = default_float
        else:
            dtype = default_int
    if stop is not None and (stop - start > 0) != (step > 0):
        return empty(0, dtype=dtype, device=device)
    if stop is None:
        return Array._new(ht.arange(0, start, step, dtype=dtype, device=device))
    else:
        return Array._new(ht.arange(start, stop, step, dtype=dtype, device=device))


def asarray(
    obj: Union[
        Array,
        bool,
        int,
        float,
        NestedSequence[bool | int | float],
        SupportsBufferProtocol,
    ],
    /,
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
    copy: Optional[bool] = None,
) -> Array:
    """
    Convert the input to an array.

    Parameters
    ----------
    obj : Union[Array, bool, int, float, NestedSequence[bool | int | float], SupportsBufferProtocol]
        Object to be converted to an array. May be a Python scalar,
        a (possibly nested) sequence of Python scalars, or an object
        supporting the Python buffer protocol. Default: ``None``.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from the data type(s) in ``obj``.
    device : Optional[Device]
        Device on which to place the created array. If ``device`` is ``None`` and
        ``x`` is an array, the output array device is inferred from ``x``.
        Default: ``None``.
    copy : Optional[bool]
        Boolean indicating whether or not to copy the input.
    """
    # _array_object imports in this file are inside the functions to avoid
    # circular imports
    from ._array_object import Array

    if isinstance(obj, Array):
        if dtype is not None and obj.dtype != dtype:
            copy = True
        if not copy:
            return obj
        obj = obj._array
    if dtype is None:
        if isinstance(obj, int) and (obj > 2**64 or obj < -(2**63)):
            raise OverflowError("Integer out of bounds for array dtypes")
        elif isinstance(obj, float):
            dtype = default_float
    res = ht.asarray(obj, dtype=dtype, device=device)
    return Array._new(res)


def empty(
    shape: Union[int, Tuple[int, ...]],
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns an uninitialized array having a specified shape.

    Parameters
    ----------
    shape : Union[int, Tuple[int, ...]]
        Output array shape.
    dtype : Optional[Dtype]
        Output array data type. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    if dtype is None:
        dtype = default_float
    return Array._new(ht.empty(shape, dtype=dtype, device=device))


def empty_like(
    x: Array, /, *, dtype: Optional[Dtype] = None, device: Optional[Device] = None
) -> Array:
    """
    Returns an uninitialized array with the same ``shape`` as an input array ``x``.

    Parameters
    ----------
    x : Array
        Input array from which to derive the output array shape.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from x. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    return Array._new(ht.empty_like(x._array, dtype=dtype, device=device))


def eye(
    n_rows: int,
    n_cols: Optional[int] = None,
    /,
    *,
    k: int = 0,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns a two-dimensional array with ones on the ``k`` h diagonal and zeros elsewhere.

    Parameters
    ----------
    n_rows : int
        Number of rows in the output array.
    n_cols : Optional[int]
        Number of columns in the output array. If ``None``, the default number of
        columns in the output array is equal to ``n_rows``. Default: ``None``.
    k : int
        Index of the diagonal. A positive value refers to an upper diagonal, a negative
        value to a lower diagonal, and ``0`` to the main diagonal. Default: ``0``.
    dtype : Optional[Dtype]
        Output array data type. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    if k != 0:
        raise ValueError("k option not implemented yet")

    if dtype is None:
        dtype = default_float
    if n_cols is None:
        n_cols = n_rows
    return Array._new(ht.eye((n_rows, n_cols), dtype=dtype, device=device))


def from_dlpack(x: object, /) -> Array:
    """
    Returns a new array containing the data from another (array) object with a
    ``__dlpack__`` method.

    Parameters
    ----------
    x : object
        Input (array) object.
    """
    from ._array_object import Array

    return Array._new(torch.from_dlpack(x))


def full(
    shape: Union[int, Tuple[int, ...]],
    fill_value: Union[int, float],
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns a new array having a specified ``shape`` and filled with ``fill_value``.

    Parameters
    ----------
    shape : Union[int, Tuple[int, ...]]
        Output array shape.
    fill_value : Union[int, float]
        Fill value.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from ``fill_value``. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array.
    """
    from ._array_object import Array

    if isinstance(fill_value, Array) and fill_value.ndim == 0:
        fill_value = fill_value._array
    if dtype is None:
        if isinstance(fill_value, bool):
            dtype = api_bool
        elif isinstance(fill_value, int):
            dtype = default_int
        elif isinstance(fill_value, float):
            dtype = default_float
    res = ht.full(shape, fill_value, dtype=dtype, device=device)
    return Array._new(res)


def full_like(
    x: Array,
    /,
    fill_value: Union[int, float],
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns a new array filled with ``fill_value`` and having the same ``shape``
    as an input array ``x``.

    Parameters
    ----------
    x : Array
        Input array from which to derive the output array shape.
    fill_value : Union[int, float]
        Fill value.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from x. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    res = ht.full_like(x._array, fill_value, dtype=dtype, device=device)
    if res.dtype not in _all_dtypes:
        # This will happen if the fill value is not something that Heat
        # coerces to one of the acceptable dtypes.
        raise TypeError("Invalid input to full_like")
    return Array._new(res)


def linspace(
    start: Union[int, float],
    stop: Union[int, float],
    /,
    num: int,
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
    endpoint: bool = True,
) -> Array:
    """
    Returns evenly spaced numbers over a specified interval.

    Parameters
    ----------
    start : Union[int, float]
        The start of the interval.
    stop : Union[int, float]
        The end of the interval.
    num : int
        Number of samples.
    dtype : Optional[Dtype]
        Output array data type. Must be a floating-point data type. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    endpoint : bool
        Boolean indicating whether to include ``stop`` in the interval. Default: ``True``.
    """
    from ._array_object import Array

    if dtype is None:
        dtype = default_float
    elif dtype not in _floating_dtypes:
        raise TypeError("Only floating dtypes allowed in linspace")

    return Array._new(ht.linspace(start, stop, num, dtype=dtype, device=device, endpoint=endpoint))


def meshgrid(*arrays: Array, indexing: str = "xy") -> List[Array]:
    """
    Returns coordinate matrices from coordinate vectors.

    Parameters
    ----------
    arrays : Array
        An arbitrary number of one-dimensional arrays representing grid coordinates.
        Each array must have the same numeric data type.
    indexing : str
        Cartesian ``'xy'`` or matrix ``'ij'`` indexing of output. If provided zero or
        one one-dimensional vector(s) (i.e., the zero- and one-dimensional cases,
        respectively), the ``indexing`` keyword has no effect and is ignored.
        Default: ``'xy'``.
    """
    from ._array_object import Array

    if len({a.dtype for a in arrays}) > 1:
        raise ValueError("meshgrid inputs must all have the same dtype")

    return [
        Array._new(array) for array in ht.meshgrid(*[a._array for a in arrays], indexing=indexing)
    ]


def ones(
    shape: Union[int, Tuple[int, ...]],
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns a new array having a specified ``shape`` and filled with ones.

    Parameters
    ----------
    shape : Union[int, Tuple[int, ...]]
        Output array shape.
    dtype : Optional[Dtype]
        Output array data type. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array.
    """
    from ._array_object import Array

    if dtype is None:
        dtype = default_float
    return Array._new(ht.ones(shape, dtype=dtype, device=device))


def ones_like(
    x: Array, /, *, dtype: Optional[Dtype] = None, device: Optional[Device] = None
) -> Array:
    """
    Returns a new array filled with ones and having the same shape as an input array ``x``.

    Parameters
    ----------
    x : Array
        Input array from which to derive the output array shape.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from x. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    return Array._new(ht.ones_like(x._array, dtype=dtype, device=device))


def tril(x: Array, /, *, k: int = 0) -> Array:
    """
    Returns the lower triangular part of a matrix (or a stack of matrices) ``x``.

    Parameters
    ----------
    x : Array
        Input array having shape ``(..., M, N)`` and whose innermost two dimensions
        form ``MxN`` matrices.
    k : int
        Diagonal above which to zero elements. If ``k = 0``, the diagonal is the
        main diagonal. If ``k < 0``, the diagonal is below the main diagonal.
        If ``k > 0``, the diagonal is above the main diagonal. Default: ``0``.
    """
    from ._array_object import Array

    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for tril")
    return Array._new(ht.tril(x._array, k=k))


def triu(x: Array, /, *, k: int = 0) -> Array:
    """
    Returns the upper triangular part of a matrix (or a stack of matrices) ``x``.

    Parameters
    ----------
    x : Array
        Input array having shape ``(..., M, N)`` and whose innermost two dimensions
        form ``MxN`` matrices.
    k : int
        Diagonal below which to zero elements. If ``k = 0``, the diagonal is the
        main diagonal. If ``k < 0``, the diagonal is below the main diagonal.
        If ``k > 0``, the diagonal is above the main diagonal. Default: ``0``.
    """
    from ._array_object import Array

    if x.ndim < 2:
        raise ValueError("x must be at least 2-dimensional for triu")
    return Array._new(ht.triu(x._array, k=k))


def zeros(
    shape: Union[int, Tuple[int, ...]],
    *,
    dtype: Optional[Dtype] = None,
    device: Optional[Device] = None,
) -> Array:
    """
    Returns a new array having a specified ``shape`` and filled with zeros.

    Parameters
    ----------
    shape : Union[int, Tuple[int, ...]]
        Output array shape.
    dtype : Optional[Dtype]
        Output array data type. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array.
    """
    from ._array_object import Array

    if dtype is None:
        dtype = default_float
    return Array._new(ht.zeros(shape, dtype=dtype, device=device))


def zeros_like(
    x: Array, /, *, dtype: Optional[Dtype] = None, device: Optional[Device] = None
) -> Array:
    """
    Returns a new array filled with zeros and having the same shape as an input array x.

    Parameters
    ----------
    x : Array
        Input array from which to derive the output array shape.
    dtype : Optional[Dtype]
        Output array data type. If ``dtype`` is ``None``, the output array data
        type is inferred from x. Default: ``None``.
    device : Optional[Device]
        Device on which to place the created array. Default: ``None``.
    """
    from ._array_object import Array

    return Array._new(ht.zeros_like(x._array, dtype=dtype))



from __future__ import annotations

from ._array_object import Array

from typing import Optional, Tuple, Union

import heat as ht


def all(
    x: Array,
    /,
    *,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
) -> Array:
    """
    Tests whether all input array elements evaluate to ``True`` along a
    specified axis.

    Parameters
    ----------
    x : Array
        Input array.
    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis or axes along which to perform a logical AND reduction. By
        default, a logical AND reduction is performed over the entire array.
    keepdims : bool
        If ``True``, the reduced axes (dimensions) are included in the result
        as singleton dimensions. Otherwise, if ``False``, the reduced axes
        (dimensions) are be included in the result. Default: ``False``.
    """
    res = ht.all(x._array, axis=axis, keepdim=keepdims)

    return Array._new(res)


def any(
    x: Array,
    /,
    *,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    keepdims: bool = False,
) -> Array:
    """
    Tests whether any input array element evaluates to ``True`` along a specified
    axis.

    Parameters
    ----------
    x : Array
        Input array.
    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis or axes along which to perform a logical OR reduction. By
        default, a logical OR reduction is performed over the entire array.
    keepdims : bool
        If ``True``, the reduced axes (dimensions) are included in the result
        as singleton dimensions. Otherwise, if ``False``, the reduced axes
        (dimensions) are be included in the result. Default: ``False``.
    """
    res = ht.any(x._array, axis=axis, keepdim=keepdims)

    return Array._new(res)



from __future__ import annotations

from ._array_object import Array
from ._data_type_functions import result_type

from typing import Optional, Tuple, Union, List

import heat as ht


def concat(arrays: Union[Tuple[Array, ...], List[Array]], /, *, axis: Optional[int] = 0) -> Array:
    """
    Joins a sequence of arrays along an existing axis.

    Parameters
    ----------
    arrays : Union[Tuple[Array, ...], List[Array]]
        Input arrays to join. The arrays must have the same shape,
        except in the dimension specified by ``axis``.
    axis : Optional[int]
        Axis along which the arrays will be joined. If ``axis`` is ``None``,
        arrays are flattened before concatenation. Default: ``0``.
    """
    result_type(*arrays)
    arrays = tuple(a._array for a in arrays)
    if axis is None:
        arrays = tuple(ht.flatten(a) for a in arrays)
        axis = 0
    return Array._new(ht.concatenate(arrays, axis=axis))


def expand_dims(x: Array, /, *, axis: int = 0) -> Array:
    """
    Expands the shape of an array by inserting a new axis (dimension) of
    size one at the position specified by ``axis``.

    Parameters
    ----------
    x : Array
        Input array.
    axis : int
        Axis position (zero-based). If ``x`` has rank (i.e, number of dimensions) ``N``,
        a valid ``axis`` must reside on the closed-interval ``[-N-1, N]``.
    """
    if axis < -x.ndim - 1 or axis > x.ndim:
        raise IndexError("Invalid axis")
    return Array._new(ht.expand_dims(x._array, axis))


def flip(x: Array, /, *, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> Array:
    """
    Reverses the order of elements in an array along the given ``axis``. The
    shape of the array is preserved.

    Parameters
    ----------
    x : Array
        Input array.
    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis (or axes) along which to flip. If ``axis`` is ``None``, the function
        flips all input array axes.
    """
    return Array._new(ht.flip(x._array, axis=axis))


def permute_dims(x: Array, /, axes: Tuple[int, ...]) -> Array:
    """
    Permutes the axes (dimensions) of an array ``x``.

    Parameters
    ----------
    x : Array
        Input array.
    axes : Tuple[int, ...]
        Tuple containing a permutation of ``(0, 1, ..., N-1)`` where ``N`` is the
        number of axes (dimensions) of ``x``.
    """
    return Array._new(ht.transpose(x._array, list(axes)))


def reshape(x: Array, /, shape: Tuple[int, ...], *, copy: Optional[bool] = None) -> Array:
    """
    Reshapes an array without changing its data.

    Parameters
    ----------
    x : Array
        Input array to reshape.
    shape : Tuple[int, ...]
        A new shape compatible with the original shape. One shape dimension
        is allowed to be ``-1``. When a shape dimension is ``-1``, the
        corresponding output array shape dimension is inferred from the length
        of the array and the remaining dimensions.
    copy : Optional[bool]
        Boolean indicating whether or not to copy the input array.
    """
    res = ht.reshape(x._array, shape)
    if not copy:
        x._array = res
        return x
    return Array._new(res)


def roll(
    x: Array,
    /,
    shift: Union[int, Tuple[int, ...]],
    *,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
) -> Array:
    """
    Rolls array elements along a specified axis. Array elements that roll
    beyond the last position are re-introduced at the first position.
    Array elements that roll beyond the first position are re-introduced
    at the last position.

    Parameters
    ----------
    x : Array
        Input array.
    shift : Union[int, Tuple[int, ...]]
        Number of places by which the elements are shifted. If ``shift`` is a
        tuple, then ``axis`` must be a tuple of the same size, and each of the
        given axes is shifted by the corresponding element in ``shift``.
        If ``shift`` is an ``int`` and ``axis`` a tuple, then the same ``shift``
        is used for all specified axes.
    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis (or axes) along which elements to shift. If ``axis`` is ``None``,
        the array is flattened, shifted, and then restored to its original shape.
        Default: ``None``.
    """
    return Array._new(ht.roll(x._array, shift, axis=axis))


def squeeze(x: Array, /, axis: Union[int, Tuple[int, ...]]) -> Array:
    """
    Removes singleton dimensions (axes) from ``x``.

    Parameters
    ----------
    x : Array
        Input array.
    axis : Union[int, Tuple[int, ...]]
        Axis (or axes) to squeeze.

    Raises
    ------
    ``ValueError``, if an axis is selected with shape entry greater than one.
    """
    return Array._new(ht.squeeze(x._array, axis=axis))


def stack(arrays: Union[Tuple[Array, ...], List[Array]], /, *, axis: int = 0) -> Array:
    """
    Joins a sequence of arrays along a new axis.

    Parameters
    ----------
    arrays : Union[Tuple[array, ...], List[array]]
        Input arrays to join. Each array must have the same shape.
    axis : int
        Axis along which the arrays will be joined.
    """
    # Call result type here just to raise on disallowed type combinations
    result_type(*arrays)
    arrays = tuple(a._array for a in arrays)
    return Array._new(ht.stack(arrays, axis=axis))
